from ._base import WorldObject
from ..utils import unpack_bitfield


class Group(WorldObject):
    """A Group is almost identical to a WorldObject. Its purpose is
    to make working with groups of objects syntactically clearer.
    """


class Scene(Group):
    """The scene is a WorldObject that represents the root of a scene graph.
    It hold certain attributes that relate to the scene, such as the background color,
    fog, and environment map. Camera's and lights can also be part of a scene.
    """


class Background(WorldObject):
    """An object representing a scene background.
    Can be e.g. a gradient, a static image or a skybox.
    """

    def __init__(self, geometry=None, material=None, **kwargs):
        if geometry is not None and material is None:
            raise TypeError("You need to instantiate using Background(None, material)")
        super().__init__(None, material, **kwargs)


class Line(WorldObject):
    """An object consisting of a line represented by a list of vertices
    (3D positions). Some materials will render the line as a continuous line,
    while other materials will consider each pair of points a segment.

    The picking info of a Line (the result of
    ``renderer.get_pick_info()``) will for most materials include
    ``vertex_index`` (int) and ``segment_coord`` (float, sub-segment coordinate).
    """


class Points(WorldObject):
    """An object consisting of points represented by vertices (3D positions).

    The picking info of a Points object (the result of
    ``renderer.get_pick_info()``) will for most materials include
    ``vertex_index`` (int) and ``point_coord`` (tuple of 2 float coordinates
    in logical pixels).
    """


class Mesh(WorldObject):
    """An object consisting of triangular faces, represented by vertices
    (3D positions) and an index that defines the connectivity.

    The picking info of a Mesh (the result of
    ``renderer.get_pick_info()``) will for most materials include
    ``instance_index`` (int), ``face_index`` (int), and ``face_coord``
    (tuple of 3 floats). The latter are the barycentric coordinates for
    each vertex of the face (with values 0..1).
    """


class Image(WorldObject):
    """An object representing a 2D image in space.

    The geometry for this object consists only of `geometry.grid`: a
    texture with the 2D data.

    The picking info of an Image (the result of ``renderer.get_pick_info()``)
    will for most materials include ``index`` (tuple of 2 int),
    and ``pixel_coord`` (tuple of float subpixel coordinates).
    """

    def _wgpu_get_pick_info(self, pick_value):
        tex = self.geometry.grid
        if hasattr(tex, "texture"):
            tex = tex.texture  # tex was a view
        # This should match with the shader
        values = unpack_bitfield(pick_value, wobject_id=20, x=22, y=22)
        x = values["x"] / 4194304 * tex.size[0] - 0.5
        y = values["y"] / 4194304 * tex.size[1] - 0.5
        ix, iy = int(x + 0.5), int(y + 0.5)
        return {
            "index": (ix, iy),
            "pixel_coord": (x - ix, y - iy),
        }


class Volume(WorldObject):
    """An object representing a 3D image in space (a volume).

    The geometry for this object consists only of `geometry.grid`: a
    texture with the 3D data.

    The picking info of a Volume (the result of ``renderer.get_pick_info()``)
    will for most materials include ``index`` (tuple of 3 int),
    and ``voxel_coord`` (tuple of float subpixel coordinates).
    """

    def _wgpu_get_pick_info(self, pick_value):
        tex = self.geometry.grid
        if hasattr(tex, "texture"):
            tex = tex.texture  # tex was a view
        # This should match with the shader
        values = unpack_bitfield(pick_value, wobject_id=20, x=14, y=14, z=14)
        texcoords_encoded = values["x"], values["y"], values["z"]
        size = tex.size
        x, y, z = [(v / 16384) * s - 0.5 for v, s in zip(texcoords_encoded, size)]
        ix, iy, iz = int(x + 0.5), int(y + 0.5), int(z + 0.5)
        return {
            "index": (ix, iy, iz),
            "voxel_coord": (x - ix, y - iy, z - iz),
        }



"""
"""

import wgpu

from ...resources import Buffer, TextureView

from ._utils import to_vertex_format, to_texture_format
from ._update import update_resource, ALTTEXFORMAT
from . import registry


def ensure_pipeline(renderer, wobject):
    """Update the GPU objects associated with the given wobject. Returns
    quickly if no changes are needed. Only this function is used by the
    renderer.
    """

    shared = renderer._shared
    blender = renderer._blender
    pipelines = renderer._wobject_pipelines
    device = shared.device

    # Get wobject_pipeline dict associated with this renderer and wobject
    wobject_pipeline = pipelines.get(wobject, {})

    # This becomes not-None if we need to update the pipeline dict
    new_pipeline_infos = None

    # Do we need to recreate the pipeline_objects?
    if wobject.rev != wobject_pipeline.get("ref", ()):
        # Create fresh wobject_pipeline
        wobject_pipeline = {"ref": wobject.rev, "renderable": False, "resources": []}
        pipelines[wobject] = wobject_pipeline
        # Create pipeline_info and collect resources
        new_pipeline_infos = create_pipeline_infos(shared, blender, wobject)
        if new_pipeline_infos:
            wobject_pipeline["renderable"] = True
            wobject_pipeline["resources"] = collect_pipeline_resources(
                shared, wobject, new_pipeline_infos
            )

    # Early exit?
    if not wobject_pipeline["renderable"]:
        return None, False

    # Check if we need to update any resources. The number of resources
    # should typically be small. We could implement a hook in the
    # resource's rev setter so we only have to check one flag ... or
    # collect all resources on a rendered .. but let's not optimize
    # prematurely.
    for kind, resource in wobject_pipeline["resources"]:
        our_version = getattr(resource, "_wgpu_" + kind, (-1, None))[0]
        if resource.rev > our_version:
            update_resource(device, resource, kind)

    # Create gpu objects?
    has_changed = bool(new_pipeline_infos)
    if has_changed:
        new_wobject_pipeline = create_pipeline_objects(
            shared, blender, wobject, new_pipeline_infos
        )
        wobject_pipeline.update(new_wobject_pipeline)

    # Set in what passes this object must render
    if has_changed:
        m = {"auto": 0, "opaque": 1, "transparent": 2, "all": 3}
        render_mask = m[wobject.render_mask]
        if not render_mask:
            render_mask = new_pipeline_infos[0].get("suggested_render_mask", 3)
        wobject_pipeline["render_mask"] = render_mask

    return wobject_pipeline, has_changed


class RenderInfo:
    """The type of object passed to each wgpu render function. Contains
    all the info that it might need:
    * wobject
    * stdinfo buffer
    * blender
    """

    def __init__(self, *, wobject, stdinfo_uniform, blender):
        self.wobject = wobject
        self.stdinfo_uniform = stdinfo_uniform
        self.blender = blender


def create_pipeline_infos(shared, blender, wobject):
    """Use the render function for this wobject and material,
    and return a list of dicts representing pipelines in an abstract way.
    These dicts can then be turned into actual pipeline objects.
    """

    # Get render function for this world object,
    # and use it to get a high-level description of pipelines.
    renderfunc = registry.get_render_function(wobject)
    if renderfunc is None:
        raise ValueError(
            f"Could not get a render function for {wobject.__class__.__name__} "
            f"with {wobject.material.__class__.__name__}"
        )

    # Prepare info for the render function
    render_info = RenderInfo(
        wobject=wobject, stdinfo_uniform=shared.stdinfo_buffer, blender=blender
    )

    # Call render function
    pipeline_infos = renderfunc(render_info)
    if not pipeline_infos:
        pipeline_infos = None
    else:
        assert isinstance(pipeline_infos, list)

    return pipeline_infos


def collect_pipeline_resources(shared, wobject, pipeline_infos):

    pipeline_resources = []  # List, because order matters

    # Collect list of resources. This way we can easily iterate
    # over dependent resource on each render call. We also set the
    # usage flag of buffers and textures specified in the pipeline.
    for pipeline_info in pipeline_infos:
        assert isinstance(pipeline_info, dict)
        buffer = pipeline_info.get("index_buffer", None)
        if buffer is not None:
            buffer._wgpu_usage |= wgpu.BufferUsage.INDEX
            pipeline_resources.append(("buffer", buffer))
        for buffer in pipeline_info.get("vertex_buffers", {}).values():
            buffer._wgpu_usage |= wgpu.BufferUsage.VERTEX
            pipeline_resources.append(("buffer", buffer))
        for key in pipeline_info.keys():
            if key.startswith("bindings"):
                resources = pipeline_info[key]
                if isinstance(resources, dict):
                    resources = resources.values()
                for binding in resources:
                    resource = binding.resource
                    if binding.type.startswith("buffer/"):
                        assert isinstance(resource, Buffer)
                        pipeline_resources.append(("buffer", resource))
                        if "uniform" in binding.type:
                            resource._wgpu_usage |= wgpu.BufferUsage.UNIFORM
                        elif "storage" in binding.type:
                            resource._wgpu_usage |= wgpu.BufferUsage.STORAGE
                    elif binding.type.startswith("sampler/"):
                        assert isinstance(resource, TextureView)
                        pipeline_resources.append(("sampler", resource))
                    elif binding.type.startswith("texture/"):
                        assert isinstance(resource, TextureView)
                        resource.texture._wgpu_usage |= (
                            wgpu.TextureUsage.TEXTURE_BINDING
                        )
                        pipeline_resources.append(("texture", resource.texture))
                        pipeline_resources.append(("texture_view", resource))
                    elif binding.type.startswith("storage_texture/"):
                        assert isinstance(resource, TextureView)
                        resource.texture._wgpu_usage |= (
                            wgpu.TextureUsage.STORAGE_BINDING
                        )
                        pipeline_resources.append(("texture", resource.texture))
                        pipeline_resources.append(("texture_view", resource))
                    else:
                        raise RuntimeError(
                            f"Unknown resource binding {binding.name} of type {binding.type}"
                        )

    return pipeline_resources


def create_pipeline_objects(shared, blender, wobject, pipeline_infos):
    """Generate wgpu pipeline objects from the list of pipeline info dicts."""

    # Prepare the three kinds of pipelines that we can get
    compute_pipelines = []
    render_pipelines = []
    alt_render_pipelines = []

    # Process each pipeline info object, converting each to a more concrete dict
    for pipeline_info in pipeline_infos:
        if "render_shader" in pipeline_info:
            pipeline = compose_render_pipeline(shared, blender, wobject, pipeline_info)
            if pipeline_info.get("target", None) is None:
                render_pipelines.append(pipeline)
            else:
                raise NotImplementedError("Alternative render pipelines")
                alt_render_pipelines.append(pipeline)
        elif "compute_shader" in pipeline_info:
            compute_pipelines.append(
                compose_compute_pipeline(shared, wobject, pipeline_info)
            )
        else:
            raise ValueError(
                "Did not find compute_shader nor render_shader in pipeline info."
            )

    return {
        "compute_pipelines": compute_pipelines,
        "render_pipelines": render_pipelines,
        "alt_render_pipelines": alt_render_pipelines,
    }


def compose_compute_pipeline(shared, wobject, pipeline_info):
    """Given a high-level compute pipeline description, creates a
    lower-level representation that can be consumed by wgpu.
    """

    # todo: cache the pipeline with the shader (and entrypoint) as a hash

    device = shared.device

    # Convert indices to args for the compute_pass.dispatch() call
    indices = pipeline_info["indices"]
    if not (
        isinstance(indices, tuple)
        and len(indices) == 3
        and all(isinstance(i, int) for i in indices)
    ):
        raise RuntimeError(f"Compute indices must be 3-tuple of ints, not {indices}.")
    index_args = indices

    # Get bind groups and pipeline layout from the buffers in pipeline_info.
    # This also makes sure the buffers and textures are up-to-date.
    bind_groups, pipeline_layout = get_bind_groups(shared, pipeline_info)

    # Compile shader and create pipeline object
    cshader = pipeline_info["compute_shader"]
    cs_module = device.create_shader_module(code=cshader)
    compute_pipeline = device.create_compute_pipeline(
        layout=pipeline_layout,
        compute={"module": cs_module, "entry_point": "main"},
    )

    return {
        "pipeline": compute_pipeline,  # wgpu object
        "index_args": index_args,  # tuple
        "bind_groups": bind_groups,  # list of wgpu bind_group objects
    }


def compose_render_pipeline(shared, blender, wobject, pipeline_info):
    """Given a high-level render pipeline description, creates a
    lower-level representation that can be consumed by wgpu.
    """

    # todo: cache the pipeline with a lot of things as the hash
    # todo: cache vertex descriptors

    device = shared.device

    # If an index buffer is present, update it, and get index_format.
    wgpu_index_buffer = None
    index_format = wgpu.IndexFormat.uint32
    index_buffer = pipeline_info.get("index_buffer", None)
    if index_buffer is not None:
        wgpu_index_buffer = index_buffer._wgpu_buffer[1]
        index_format = to_vertex_format(index_buffer.format)
        index_format = index_format.split("x")[0].replace("s", "u")

    # Convert and check high-level indices. Indices represent a range
    # of index id's, or define what indices in the index buffer are used.
    indices = pipeline_info.get("indices", None)
    if indices is None:
        if index_buffer is None:
            raise RuntimeError("Need indices or index_buffer ")
        indices = range(index_buffer.data.size)
    # Convert to 2-element tuple (vertex, instance)
    if not isinstance(indices, tuple):
        indices = (indices,)
    if len(indices) == 1:
        indices = indices + (1,)  # add instancing index
    if len(indices) != 2:
        raise RuntimeError("Render pipeline indices must be a 2-element tuple.")

    # Convert indices to args for the render_pass.draw() or draw_indexed()
    # draw(count_vertex, count_instance, first_vertex, first_instance)
    # draw_indexed(count_v, count_i, first_vertex, base_vertex, first_instance)
    index_args = [0, 0, 0, 0]
    for i, index in enumerate(indices):
        if isinstance(index, int):
            index_args[i] = index
        elif isinstance(index, range):
            assert index.step == 1
            index_args[i] = index.stop - index.start
            index_args[i + 2] = index.start
        else:
            raise RuntimeError(
                "Render pipeline indices must be a 2-element tuple with ints or ranges."
            )
    if wgpu_index_buffer is not None:
        base_vertex = 0  # A value added to each index before reading [...]
        index_args.insert(-1, base_vertex)

    # Process vertex buffers. Update the buffer, and produces a descriptor.
    vertex_buffers = {}
    vertex_buffer_descriptors = []
    # todo: we can probably expose multiple attributes per buffer using a BufferView
    # -> can we also leverage numpy here?
    for slot, buffer in pipeline_info.get("vertex_buffers", {}).items():
        slot = int(slot)
        vbo_des = {
            "array_stride": buffer.nbytes // buffer.nitems,
            "step_mode": wgpu.VertexStepMode.vertex,  # vertex or instance
            "attributes": [
                {
                    "format": to_vertex_format(buffer.format),
                    "offset": 0,
                    "shader_location": slot,
                }
            ],
        }
        vertex_buffers[slot] = buffer
        vertex_buffer_descriptors.append(vbo_des)

    # Get bind groups and pipeline layout from the buffers in pipeline_info.
    # This also makes sure the buffers and textures are up-to-date.
    bind_groups, pipeline_layout = get_bind_groups(shared, pipeline_info)

    # todo: is this how strip_index_format is supposed to work?
    strip_index_format = 0
    if "strip" in pipeline_info["primitive_topology"]:
        strip_index_format = index_format

    # Instantiate the pipeline objects

    pipelines = {}
    for pass_index in range(blender.get_pass_count()):
        color_descriptors = blender.get_color_descriptors(pass_index)
        depth_descriptor = blender.get_depth_descriptor(pass_index)
        if not color_descriptors:
            continue

        # Compile shader
        shader = pipeline_info["render_shader"]
        wgsl = shader.generate_wgsl(**blender.get_shader_kwargs(pass_index))
        shader_module = get_shader_module(shared, wgsl)

        pipelines[pass_index] = device.create_render_pipeline(
            layout=pipeline_layout,
            vertex={
                "module": shader_module,
                "entry_point": "vs_main",
                "buffers": vertex_buffer_descriptors,
            },
            primitive={
                "topology": pipeline_info["primitive_topology"],
                "strip_index_format": strip_index_format,
                "front_face": wgpu.FrontFace.ccw,
                "cull_mode": pipeline_info.get("cull_mode", wgpu.CullMode.none),
            },
            depth_stencil={
                **depth_descriptor,
                "stencil_front": {},  # use defaults
                "stencil_back": {},  # use defaults
            },
            multisample={
                "count": 1,
                "mask": 0xFFFFFFFF,
                "alpha_to_coverage_enabled": False,
            },
            fragment={
                "module": shader_module,
                "entry_point": "fs_main",
                "targets": color_descriptors,
            },
        )

    return {
        "pipelines": pipelines,  # wgpu objects
        "index_args": index_args,  # tuple
        "index_buffer": wgpu_index_buffer,  # Buffer
        "vertex_buffers": vertex_buffers,  # dict of slot -> Buffer
        "bind_groups": bind_groups,  # list of wgpu bind_group objects
    }


def get_bind_groups(shared, pipeline_info):
    """Given high-level information on bindings, create the corresponding
    wgpu objects. This assumes that all buffers and textures are up-to-date.
    Returns (bind_groups, pipeline_layout).
    """
    # todo: cache bind_group_layout objects
    # todo: cache pipeline_layout objects

    device = shared.device

    # Collect resource groups (keys e.g. "bindings1", "bindings132")
    resource_groups = []
    for key in pipeline_info.keys():
        if key.startswith("bindings"):
            i = int(key[len("bindings") :])
            assert i >= 0
            if not pipeline_info[key]:
                continue
            while len(resource_groups) <= i:
                resource_groups.append({})
            resource_groups[i] = pipeline_info[key]

    # Create bind groups and bind group layouts
    bind_groups = []
    bind_group_layouts = []
    for resources in resource_groups:
        if not isinstance(resources, dict):
            resources = {slot: resource for slot, resource in enumerate(resources)}
        # Collect list of dicts
        bindings = []
        binding_layouts = []
        for slot, binding in resources.items():
            resource = binding.resource
            subtype = binding.type.partition("/")[2]

            if binding.type.startswith("buffer/"):
                assert isinstance(resource, Buffer)
                bindings.append(
                    {
                        "binding": slot,
                        "resource": {
                            "buffer": resource._wgpu_buffer[1],
                            "offset": 0,
                            "size": resource.nbytes,
                        },
                    }
                )
                binding_layouts.append(
                    {
                        "binding": slot,
                        "visibility": binding.visibility,
                        "buffer": {
                            "type": getattr(wgpu.BufferBindingType, subtype),
                            "has_dynamic_offset": False,
                            "min_binding_size": 0,
                        },
                    }
                )
            elif binding.type.startswith("sampler/"):
                assert isinstance(resource, TextureView)
                bindings.append(
                    {"binding": slot, "resource": resource._wgpu_sampler[1]}
                )
                binding_layouts.append(
                    {
                        "binding": slot,
                        "visibility": binding.visibility,
                        "sampler": {
                            "type": getattr(wgpu.SamplerBindingType, subtype),
                        },
                    }
                )
            elif binding.type.startswith("texture/"):
                assert isinstance(resource, TextureView)
                bindings.append(
                    {"binding": slot, "resource": resource._wgpu_texture_view[1]}
                )
                dim = resource.view_dim
                dim = getattr(wgpu.TextureViewDimension, dim, dim)
                sample_type = getattr(wgpu.TextureSampleType, subtype, subtype)
                # Derive sample type from texture
                if sample_type == "auto":
                    fmt = to_texture_format(resource.format)
                    fmt = ALTTEXFORMAT.get(fmt, [fmt])[0]
                    if "float" in fmt or "norm" in fmt:
                        sample_type = wgpu.TextureSampleType.float
                        # For float32 wgpu does not allow the sampler to be filterable,
                        # except when the native-only feature
                        # TEXTURE_ADAPTER_SPECIFIC_FORMAT_FEATURES is set,
                        # which wgpu-py does by default.
                        # if "32float" in fmt:
                        #     sample_type = wgpu.TextureSampleType.unfilterable_float
                    elif "uint" in fmt:
                        sample_type = wgpu.TextureSampleType.uint
                    elif "sint" in fmt:
                        sample_type = wgpu.TextureSampleType.sint
                    elif "depth" in fmt:
                        sample_type = wgpu.TextureSampleType.depth
                    else:
                        raise ValueError("Could not determine texture sample type.")
                binding_layouts.append(
                    {
                        "binding": slot,
                        "visibility": binding.visibility,
                        "texture": {
                            "sample_type": sample_type,
                            "view_dimension": dim,
                            "multisampled": False,
                        },
                    }
                )
            elif binding.type.startswith("storage_texture/"):
                assert isinstance(resource, TextureView)
                bindings.append(
                    {"binding": slot, "resource": resource._wgpu_texture_view[1]}
                )
                dim = resource.view_dim
                dim = getattr(wgpu.TextureViewDimension, dim, dim)
                fmt = to_texture_format(resource.format)
                fmt = ALTTEXFORMAT.get(fmt, [fmt])[0]
                binding_layouts.append(
                    {
                        "binding": slot,
                        "visibility": binding.visibility,
                        "storage_texture": {
                            "access": getattr(wgpu.StorageTextureAccess, subtype),
                            "format": fmt,
                            "view_dimension": dim,
                        },
                    }
                )

        # Create wgpu objects
        bind_group_layout = device.create_bind_group_layout(entries=binding_layouts)
        bind_group = device.create_bind_group(
            layout=bind_group_layout, entries=bindings
        )
        bind_groups.append(bind_group)
        bind_group_layouts.append(bind_group_layout)

    # Create pipeline layout object from list of layouts
    pipeline_layout = device.create_pipeline_layout(
        bind_group_layouts=bind_group_layouts
    )

    return bind_groups, pipeline_layout


def get_shader_module(shared, source):
    """Compile a shader module object, or re-use it from the cache."""
    # todo: also release shader modules that are no longer used

    assert isinstance(source, str)
    key = source  # or hash(code)

    # AK: this dev snippet is so useful that I leave it as a comment.
    # After a crash you can run naga on the generated file to get better feedback.
    # with open(__import__("os").path.expanduser("~/tmp.wgsl"), "wb") as f:
    #     f.write(source.encode())

    if key not in shared.shader_cache:
        m = shared.device.create_shader_module(code=source)
        shared.shader_cache[key] = m
    return shared.shader_cache[key]



import numpy as np

from ._base import Geometry


def trimesh_geometry(mesh):
    """Create geometry from a trimesh Mesh."""
    kwargs = dict(
        positions=np.ascontiguousarray(mesh.vertices, dtype="f4"),
        indices=np.ascontiguousarray(mesh.faces, dtype="i4"),
        normals=np.ascontiguousarray(mesh.vertex_normals, dtype="f4"),
    )
    if mesh.visual.kind == "texture":
        kwargs["texcoords"] = np.ascontiguousarray(mesh.visual.uv, dtype="f4")
    elif mesh.visual.kind == "vertex":
        kwargs["colors"] = np.ascontiguousarray(mesh.visual.vertex_colors, dtype="f4")
    return Geometry(**kwargs)



# flake8: noqa

"""
Controllers define how cameras can be interacted with. They are
independent of any GUI toolkits.
"""

from ._base import Controller
from ._orbit import OrbitController, OrbitOrthoController
from ._panzoom import PanZoomController


__all__ = ["Controller", "PanZoomController", "OrbitController", "OrbitOrthoController"]



import numpy as np

from ..objects._base import ResourceContainer
from ..resources import Resource, Buffer, Texture
from ..linalg.utils import aabb_to_sphere


class Geometry(ResourceContainer):
    """A Geomerty object is a container for geometry data of a WorldObject.

    A geometry object contains the data that defines (the shape of) the
    object, such as positions, plus data associated with these positions
    (normals, texcoords, colors, etc.).

    Its attributes are Buffer and Texture(View) objects. The data can
    be provided as keyword arguments, which are converted to numpy
    arrays and wrapped in a Buffer if necessary.

    The names for these attributes are standardized so that the
    renderers know what to expect. Each material requires certain
    attributes to be present, and may support optional attributes.
    Optional attributes must always be "turned on" on the material;
    their presence on the geometry does not mean that they're used.

    The standardized names are:

    * ``indices``: An index into per-vertex data. Typically Nx3 for mesh geometry.
    * ``positions``: Nx3 positions (xyz), defining the location of e.g. vertices or points.
    * ``normals``: Nx3 normal vectors. These may or may not be unit.
    * ``texcoords``: Texture coordinates used to lookup the color for a vertex.
      Can be Nx1, Nx2 or Nx3, corresponding to a 1D, 2D and 3D texture map.
    * ``colors``: Per-vertex colors. Must be NxM, with M 1-4 for gray,
      gray+alpha, rgb, rgba, respectively.
    * ``sizes``: Scalar size per-vertex.
    * ``grid``: A 2D or 3D Texture/TextureView that contains a regular grid of
      data. I.e. for images and volumes.

    :Example:

    .. code-block:: py

        g = Geometry(positions=[[1, 2], [2, 4], [3, 5], [4, 1]])
        g.positions.data  # numpy array

    """

    def __init__(self, **kwargs):
        super().__init__()

        self._aabb = None
        self._aabb_rev = None
        self._bsphere = None
        self._bsphere_rev = None

        for name, val in kwargs.items():

            # Get resource object
            if isinstance(val, Resource):
                resource = val
            else:
                if not isinstance(val, np.ndarray):
                    dtype = np.uint32 if name == "indices" else np.float32
                    val = np.asanyarray(val, dtype=dtype)
                if val.dtype == np.float64:
                    raise ValueError(
                        "64-bit float is not supported, use 32-bit floats instead"
                    )
                if name == "grid":
                    dim = val.ndim
                    if dim > 2 and val.shape[-1] <= 4:
                        dim -= 1  # last array dim is probably (a subset of) rgba
                    resource = Texture(val, dim=dim).get_view()
                else:
                    resource = Buffer(val)

            # Checks
            if isinstance(resource, Buffer):
                format = resource.format
                if name == "indices":
                    pass  # No assumptions about shape; they're considered flat anyway
                elif name == "positions":
                    if not format.startswith("3x"):
                        raise ValueError("Expected Nx3 data for positions")
                elif name == "normals":
                    if not format.startswith("3x"):
                        raise ValueError("Expected Nx3 data for normals")
                elif name == "texcoords":
                    if not ("x" not in format or format.startswith(("1x", "2x", "3x"))):
                        raise ValueError("Expected Nx1, Nx2 or Nx3 data for texcoords")
                elif name == "colors":
                    if not format.startswith("4x"):
                        raise ValueError("Expected Nx4 data for colors")
                elif name == "sizes":
                    if not ("x" not in format):
                        raise ValueError("Expected array of scalars for sizes")
                else:
                    pass  # Unknown attribute - no checks

            # Store
            setattr(self, name, resource)

    def bounding_box(self):
        """Compute the axis-aligned bounding box based on either positions
        or the shape of the grid buffer.

        If both are present, the bounding box will be computed based on
        the positions buffer.
        """
        if hasattr(self, "positions"):
            if self._aabb_rev == self.positions.rev:
                return self._aabb
            pos = self.positions.data
            self._aabb = np.array([pos.min(axis=0), pos.max(axis=0)])
            self._aabb_rev = self.positions.rev
            return self._aabb

        if hasattr(self, "grid"):
            if self._aabb_rev == self.grid.rev:
                return self._aabb
            self._aabb = (
                np.array([np.zeros_like(self.grid.data.shape), self.grid.data.shape])
                - 0.5
            )
            self._aabb_rev = self.grid.rev
            return self._aabb

        raise ValueError(
            "No positions or grid buffer available for bounding volume computation"
        )

    def bounding_sphere(self):
        """Compute the bounding sphere based on the axis-aligned bounding box.

        Note: not the optimal fit.
        """
        if self._bsphere is not None and self._bsphere_rev == self._aabb_rev:
            return self._bsphere

        self._bsphere = aabb_to_sphere(self.bounding_box())
        self._bsphere_rev = self._aabb_rev
        return self._bsphere



import wgpu  # only for flags/enums

from . import register_wgpu_render_function
from ._shadercomposer import Binding, WorldObjectShader
from ._utils import to_vertex_format, to_texture_format
from ...objects import Points
from ...materials import PointsMaterial, GaussianPointsMaterial
from ...resources import Texture, TextureView


def handle_colormap(geometry, material, shader):
    if isinstance(material.map, Texture):
        raise TypeError("material.map is a Texture, but must be a TextureView")
    elif not isinstance(material.map, TextureView):
        raise TypeError("material.map must be a TextureView")
    elif getattr(geometry, "texcoords", None) is None:
        raise ValueError("material.map is present, but geometry has no texcoords")
    # Dimensionality
    shader["colormap_dim"] = view_dim = material.map.view_dim
    if view_dim not in ("1d", "2d", "3d"):
        raise ValueError("Unexpected texture dimension")
    # Texture dim matches texcoords
    vert_fmt = to_vertex_format(geometry.texcoords.format)
    if view_dim == "1d" and "x" not in vert_fmt:
        pass
    elif not vert_fmt.endswith("x" + view_dim[0]):
        raise ValueError(
            f"geometry.texcoords {geometry.texcoords.format} does not match material.map {view_dim}"
        )
    # Sampling type
    fmt = to_texture_format(material.map.format)
    if "norm" in fmt or "float" in fmt:
        shader["colormap_format"] = "f32"
    elif "uint" in fmt:
        shader["colormap_format"] = "u32"
    else:
        shader["colormap_format"] = "i32"
    # Channels
    shader["colormap_nchannels"] = len(fmt) - len(fmt.lstrip("rgba"))
    # Return bindings
    return [
        Binding("s_colormap", "sampler/filtering", material.map, "FRAGMENT"),
        Binding("t_colormap", "texture/auto", material.map, "FRAGMENT"),
        Binding(
            "s_texcoords", "buffer/read_only_storage", geometry.texcoords, "VERTEX"
        ),
    ]


@register_wgpu_render_function(Points, PointsMaterial)
def points_renderer(render_info):
    """Render function capable of rendering Points."""

    wobject = render_info.wobject
    geometry = wobject.geometry
    material = wobject.material
    shader = PointsShader(
        render_info,
        type="circle",
        per_vertex_sizes=False,
        vertex_color_channels=0,
    )
    n = geometry.positions.nitems * 6

    bindings = [
        Binding("u_stdinfo", "buffer/uniform", render_info.stdinfo_uniform),
        Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
        Binding("u_material", "buffer/uniform", material.uniform_buffer),
        Binding(
            "s_positions", "buffer/read_only_storage", geometry.positions, "VERTEX"
        ),
    ]

    if material.vertex_sizes:
        shader["per_vertex_sizes"] = True
        bindings.append(
            Binding("s_sizes", "buffer/read_only_storage", geometry.sizes, "VERTEX")
        )

    # Per-vertex color, colormap, or a plane color?
    shader["color_mode"] = "uniform"
    if material.vertex_colors:
        shader["color_mode"] = "vertex"
        shader["vertex_color_channels"] = nchannels = geometry.colors.data.shape[1]
        if nchannels not in (1, 2, 3, 4):
            raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
        bindings.append(
            Binding("s_colors", "buffer/read_only_storage", geometry.colors, "VERTEX")
        )
    elif material.map is not None:
        shader["color_mode"] = "map"
        bindings.extend(handle_colormap(geometry, material, shader))

    if isinstance(material, GaussianPointsMaterial):
        shader["type"] = "gaussian"

    # Let the shader generate code for our bindings
    for i, binding in enumerate(bindings):
        shader.define_binding(0, i, binding)

    # Determine in what render passes this objects must be rendered
    suggested_render_mask = 3
    if material.opacity < 1:
        suggested_render_mask = 2
    elif shader["color_mode"] == "vertex":
        if shader["vertex_color_channels"] in (1, 3):
            suggested_render_mask = 1
    elif shader["color_mode"] == "map":
        if shader["colormap_nchannels"] in (1, 3):
            suggested_render_mask = 1
    elif shader["color_mode"] == "uniform":
        suggested_render_mask = 1 if material.color[3] >= 1 else 2
    else:
        raise RuntimeError(f"Unexpected color mode {shader['color_mode']}")

    # Put it together!
    return [
        {
            "suggested_render_mask": suggested_render_mask,
            "render_shader": shader,
            "primitive_topology": wgpu.PrimitiveTopology.triangle_list,
            "indices": (range(n), range(1)),
            "vertex_buffers": {},
            "bindings0": bindings,
        }
    ]


class PointsShader(WorldObjectShader):

    # Notes:
    # In WGPU, the pointsize attribute can no longer be larger than 1 because
    # of restriction in some hardware/backend API's. So we use our storage-buffer
    # approach (similar for what we use for lines) to sort of fake a geometry shader.
    # An alternative is to use instancing. Could be worth testing both approaches
    # for performance ...

    def get_code(self):
        return (
            self.get_definitions()
            + self.common_functions()
            + self.vertex_shader()
            + self.fragment_shader()
        )

    def vertex_shader(self):
        return """

        struct VertexInput {
            [[builtin(vertex_index)]] vertex_index : u32;
        };


        [[stage(vertex)]]
        fn vs_main(in: VertexInput) -> Varyings {

            let index = i32(in.vertex_index);
            let i0 = index / 6;
            let sub_index = index % 6;

            let raw_pos = load_s_positions(i0);
            let world_pos = u_wobject.world_transform * vec4<f32>(raw_pos, 1.0);
            let ndc_pos = u_stdinfo.projection_transform * u_stdinfo.cam_transform * world_pos;

            var deltas = array<vec2<f32>, 6>(
                vec2<f32>(-1.0, -1.0),
                vec2<f32>(-1.0,  1.0),
                vec2<f32>( 1.0, -1.0),
                vec2<f32>(-1.0,  1.0),
                vec2<f32>( 1.0, -1.0),
                vec2<f32>( 1.0,  1.0),
            );

            // Need size here in vertex shader too
            $$ if per_vertex_sizes
                let size = load_s_sizes(i0);
            $$ else
                let size = u_material.size;
            $$ endif

            let aa_margin = 1.0;
            let delta_logical = deltas[sub_index] * (size + aa_margin);
            let delta_ndc = delta_logical * (1.0 / u_stdinfo.logical_size);

            var varyings: Varyings;
            varyings.position = vec4<f32>(ndc_pos.xy + delta_ndc * ndc_pos.w, ndc_pos.zw);
            varyings.world_pos = vec3<f32>(world_pos.xyz / world_pos.w);
            varyings.pointcoord = vec2<f32>(delta_logical);
            varyings.size = f32(size);

            // Picking
            varyings.pick_idx = u32(i0);

            // Per-vertex colors
            $$ if vertex_color_channels == 1
            let cvalue = load_s_colors(i0);
            varyings.color = vec4<f32>(cvalue, cvalue, cvalue, 1.0);
            $$ elif vertex_color_channels == 2
            let cvalue = load_s_colors(i0);
            varyings.color = vec4<f32>(cvalue.r, cvalue.r, cvalue.r, cvalue.g);
            $$ elif vertex_color_channels == 3
            varyings.color = vec4<f32>(load_s_colors(i0), 1.0);
            $$ elif vertex_color_channels == 4
            varyings.color = vec4<f32>(load_s_colors(i0));
            $$ endif

            // Set texture coords
            $$ if colormap_dim == '1d'
            varyings.texcoord = f32(load_s_texcoords(i0));
            $$ elif colormap_dim == '2d'
            varyings.texcoord = vec2<f32>(load_s_texcoords(i0));
            $$ elif colormap_dim == '3d'
            varyings.texcoord = vec3<f32>(load_s_texcoords(i0));
            $$ endif

            return varyings;
        }
        """

    def fragment_shader(self):
        # Also see See https://github.com/vispy/vispy/blob/master/vispy/visuals/markers.py
        return """

        [[stage(fragment)]]
        fn fs_main(varyings: Varyings) -> FragmentOutput {
            var final_color : vec4<f32>;

            let d = length(varyings.pointcoord);
            let aa_width = 1.0;

            $$ if per_vertex_sizes
                let size = varyings.size;
            $$ else
                let size = u_material.size;
            $$ endif

            $$ if color_mode == 'vertex'
                let color = varyings.color;
            $$ elif color_mode == 'map'
                let color = sample_colormap(varyings.texcoord);
            $$ else
                let color = u_material.color;
            $$ endif

            $$ if type == 'circle'
                if (d <= size - 0.5 * aa_width) {
                    final_color = color;
                } elseif (d <= size + 0.5 * aa_width) {
                    let alpha1 = 0.5 + (size - d) / aa_width;
                    let alpha2 = pow(alpha1, 2.0);  // this works better
                    final_color = vec4<f32>(color.rgb, color.a * alpha2);
                } else {
                    discard;
                }
            $$ elif type == "gaussian"
                if (d <= size) {
                    let sigma = size / 3.0;
                    let t = d / sigma;
                    let a = exp(-0.5 * t * t);
                    final_color = vec4<f32>(color.rgb, color.a * a);
                } else {
                    discard;
                }
            $$ else
                invalid_point_type;
            $$ endif

            final_color.a = final_color.a * u_material.opacity;

            // Wrap up
            apply_clipping_planes(varyings.world_pos);
            var out = get_fragment_output(varyings.position.z, final_color);

            $$ if write_pick
            // The wobject-id must be 20 bits. In total it must not exceed 64 bits.
            out.pick = (
                pick_pack(u32(u_wobject.id), 20) +
                pick_pack(varyings.pick_idx, 26) +
                pick_pack(u32(varyings.pointcoord.x + 256.0), 9) +
                pick_pack(u32(varyings.pointcoord.y + 256.0), 9)
            );
            $$ endif

            return out;
        }
        """



import wgpu  # only for flags/enums

from . import register_wgpu_render_function
from ._shadercomposer import Binding, WorldObjectShader
from ._utils import to_texture_format
from ...objects import Background
from ...materials import BackgroundMaterial, BackgroundImageMaterial
from ...resources import Texture, TextureView


@register_wgpu_render_function(Background, BackgroundMaterial)
def background_renderer(render_info):

    wobject = render_info.wobject
    material = wobject.material
    shader = BackgroundShader(render_info, texture_dim="")

    bindings = {}

    # Uniforms
    bindings[0] = Binding("u_stdinfo", "buffer/uniform", render_info.stdinfo_uniform)
    bindings[1] = Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer)
    bindings[2] = Binding("u_material", "buffer/uniform", material.uniform_buffer)

    if isinstance(material, BackgroundImageMaterial) and material.map is not None:
        if isinstance(material.map, Texture):
            raise TypeError("material.map is a Texture, but must be a TextureView")
        elif not isinstance(material.map, TextureView):
            raise TypeError("material.map must be a TextureView")
        bindings[3] = Binding(
            "r_sampler", "sampler/filtering", material.map, "FRAGMENT"
        )
        bindings[4] = Binding("r_tex", "texture/auto", material.map, "FRAGMENT")
        # Select shader
        if material.map.view_dim == "cube":
            shader["texture_dim"] = "cube"
        elif material.map.view_dim == "2d":
            shader["texture_dim"] = "2d"
        else:
            raise ValueError(
                "BackgroundImageMaterial should have map with texture view 2d or cube."
            )
        # Channels
        fmt = to_texture_format(material.map.format)
        shader["texture_nchannels"] = len(fmt) - len(fmt.lstrip("rgba"))

    # Let the shader generate code for our bindings
    for i, binding in bindings.items():
        shader.define_binding(0, i, binding)

    return [
        {
            "suggested_render_mask": 1,
            "render_shader": shader,
            "primitive_topology": wgpu.PrimitiveTopology.triangle_strip,
            "indices": 4,
            "bindings0": bindings,
        }
    ]


class BackgroundShader(WorldObjectShader):
    def get_code(self):
        return (
            self.get_definitions()
            + self.common_functions()
            + self.vertex_shader()
            + self.fragment_shader()
        )

    def vertex_shader(self):
        return """

        struct VertexInput {
            [[builtin(vertex_index)]] index : u32;
        };


        [[stage(vertex)]]
        fn vs_main(in: VertexInput) -> Varyings {
            var varyings: Varyings;
            // Define positions at the four corners of the viewport, at the largest depth
            var positions = array<vec2<f32>, 4>(
                vec2<f32>(-1.0, -1.0),
                vec2<f32>( 1.0, -1.0),
                vec2<f32>(-1.0,  1.0),
                vec2<f32>( 1.0,  1.0),
            );
            // Select the current position
            let pos = positions[i32(in.index)];
            $$ if texture_dim == "cube"
                let ndc_pos1 = vec4<f32>(pos, 0.9999999, 1.0);
                let ndc_pos2 = vec4<f32>(pos, 1.1000000, 1.0);
                let wpos1 = ndc_to_world_pos(ndc_pos1);
                let wpos2 = ndc_to_world_pos(ndc_pos2);
                // Store positions and the view direction in the world
                varyings.position = vec4<f32>(ndc_pos1);
                varyings.world_pos = vec3<f32>(wpos1);
                varyings.texcoord = vec3<f32>(wpos2.xyz - wpos1.xyz);
            $$ else
                // Store positions and the view direction in the world
                varyings.position = vec4<f32>(pos, 0.9999999, 1.0);
                varyings.world_pos = vec3<f32>(ndc_to_world_pos(out.position));
                varyings.texcoord = vec3<f32>(pos * 0.5 + 0.5, 0.0);
            $$ endif
            return varyings;
        }
        """

    def fragment_shader(self):
        return """
        [[stage(fragment)]]
        fn fs_main(varyings: Varyings) -> FragmentOutput {
            var final_color : vec4<f32>;
            $$ if texture_dim
                $$ if texture_dim == '2d'
                    let color = textureSample(r_tex, r_sampler, varyings.texcoord.xy);
                $$ elif texture_dim == 'cube'
                    let color = textureSample(r_tex, r_sampler, varyings.texcoord.xyz);
                $$ endif
                $$ if texture_nchannels == 1
                    final_color = vec4<f32>(color.rrr, 1.0);
                $$ elif texture_nchannels == 2
                    final_color = vec4<f32>(color.rrr, color.g);
                $$ else
                    final_color = color;
                $$ endif
            $$ else
                let f = varyings.texcoord.xy;
                final_color = (
                    u_material.color_bottom_left * (1.0 - f.x) * (1.0 - f.y)
                    + u_material.color_bottom_right * f.x * (1.0 - f.y)
                    + u_material.color_top_left * (1.0 - f.x) * f.y
                    + u_material.color_top_right * f.x * f.y
                );
            $$ endif
            final_color.a = final_color.a * u_material.opacity;

            // We can apply clipping planes, but maybe a background should not be clipped?
            // apply_clipping_planes(in.world_pos);

            $$ if not write_pick
                if (true) { discard; }
            $$ endif

            // This is the opaque pass.
            // A fragment of the background could be transparent, but it should still be
            // written in the opaque pass in order for it to really be background.
            // So we fool the blender into thinking this fragment is opaque, even if its not.
            var out = get_fragment_output(varyings.position.z, vec4<f32>(final_color.rgb, 1.0));
            out.color = final_color;
            return out;
        }
        """



# flake8: noqa

from ._utils import registry, register_wgpu_render_function
from ._renderer import stdinfo_uniform_type
from ._renderer import WgpuRenderer
from . import meshrender
from . import linerender
from . import pointsrender
from . import imagerender
from . import volumerender
from . import backgroundrender



import time
import weakref
import logging

import numpy as np
import wgpu.backends.rs

from .. import Renderer
from ...linalg import Matrix4, Vector3
from ...objects import (
    id_provider,
    KeyboardEvent,
    RootEventHandler,
    PointerEvent,
    WheelEvent,
    WindowEvent,
    WorldObject,
)
from ...cameras import Camera
from ...resources import Buffer, Texture, TextureView
from ...utils import array_from_shadertype, Color

from . import _blender as blender_module
from ._flusher import RenderFlusher
from ._pipelinebuilder import ensure_pipeline
from ._update import update_buffer, update_texture, update_texture_view


logger = logging.getLogger("pygfx")


# Definition uniform struct with standard info related to transforms,
# provided to each shader as uniform at slot 0.
# todo: a combined transform would be nice too, for performance
# todo: same for ndc_to_world transform (combined inv transforms)
stdinfo_uniform_type = dict(
    cam_transform="4x4xf4",
    cam_transform_inv="4x4xf4",
    projection_transform="4x4xf4",
    projection_transform_inv="4x4xf4",
    physical_size="2xf4",
    logical_size="2xf4",
    flipped_winding="i4",  # A bool, really
)


def _get_sort_function(camera: Camera):
    """Given a scene object, get a function to sort wobject-tuples"""

    def sort_func(wobject_tuple: WorldObject):
        wobject = wobject_tuple[0]
        z = (
            Vector3()
            .set_from_matrix_position(wobject.matrix_world)
            .apply_matrix4(proj_screen_matrix)
            .z
        )
        return wobject.render_order, z

    proj_screen_matrix = Matrix4().multiply_matrices(
        camera.projection_matrix, camera.matrix_world_inverse
    )

    return sort_func


class SharedData:
    """An object to store global data to share between multiple wgpu renderers.

    Since renderers don't render simultaneously, they can share certain
    resources. This safes memory, but more importantly, resources that
    get used in wobject pipelines should be shared to avoid having to
    constantly recompose the pipelines of wobjects that are rendered by
    multiple renderers.
    """

    def __init__(self, canvas):

        # Create adapter and device objects - there should be just one per canvas.
        # Having a global device provides the benefit that we can draw any object
        # anywhere.
        # We could pass the canvas to request_adapter(), so we get an adapter that is
        # at least compatible with the first canvas that a renderer is create for.
        # However, passing the object has been shown to prevent the creation of
        # a canvas (on Linux + wx), so, we never pass it for now.
        self.adapter = wgpu.request_adapter(
            canvas=None, power_preference="high-performance"
        )
        self.device = self.adapter.request_device(
            required_features=[], required_limits={}
        )

        # Create a uniform buffer for std info
        self.stdinfo_buffer = Buffer(array_from_shadertype(stdinfo_uniform_type))
        self.stdinfo_buffer._wgpu_usage |= wgpu.BufferUsage.UNIFORM

        # A cache for shader objects
        self.shader_cache = {}


class WgpuRenderer(RootEventHandler, Renderer):
    """Object used to render scenes using wgpu.

    The purpose of a renderer is to render (i.e. draw) a scene to a
    canvas or texture. It also provides picking, defines the
    anti-aliasing parameters, and any post processing effects.

    A renderer is directly associated with its target and can only render
    to that target. Different renderers can render to the same target though.

    It provides a ``.render()`` method that can be called one or more
    times to render scenes. This creates a visual representation that
    is stored internally, and is finally rendered into its render target
    (the canvas or texture).
                                  __________
                                 | blender  |
        [scenes] -- render() --> |  state   | -- flush() --> [target]
                                 |__________|

    The internal representation is managed by the blender object. The
    internal render textures are typically at a higher resolution to
    reduce aliasing (SSAA). The blender has auxilary buffers such as a
    depth buffer, pick buffer, and buffers for transparent fragments.
    Depending on the blend mode, a single render call may consist of
    multiple passes (to deal with semi-transparent fragments).

    The flush-step resolves the internal representation into the target
    texture or canvas, averaging neighbouring fragments for anti-aliasing.

    Parameters:
        target (WgpuCanvas or Texture): The target to render to, and what
            determines the size of the render buffer.
        pixel_ratio (float, optional): How large the physical size of the render
            buffer is in relation to the target's physical size, for antialiasing.
            See the corresponding property for details.
        show_fps (bool): Whether to display the frames per second. Beware that
            depending on the GUI toolkit, the canvas may impose a frame rate limit.
    """

    _shared = None

    _wobject_pipelines_collection = weakref.WeakValueDictionary()

    def __init__(
        self,
        target,
        *args,
        pixel_ratio=None,
        show_fps=False,
        blend_mode="default",
        sort_objects=False,
        enable_events=True,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        # Check and normalize inputs
        if not isinstance(target, (Texture, TextureView, wgpu.gui.WgpuCanvasBase)):
            raise TypeError(
                f"Render target must be a canvas or texture (view), not a {target.__class__.__name__}"
            )
        self._target = target

        # Process other inputs
        self.pixel_ratio = pixel_ratio
        self._show_fps = bool(show_fps)

        # Make sure we have a shared object (the first renderer create it)
        canvas = target if isinstance(target, wgpu.gui.WgpuCanvasBase) else None
        if WgpuRenderer._shared is None:
            WgpuRenderer._shared = SharedData(canvas)

        # Init counter to auto-clear
        self._renders_since_last_flush = 0

        # Get target format
        if isinstance(target, wgpu.gui.WgpuCanvasBase):
            self._canvas_context = self._target.get_context()
            self._target_tex_format = self._canvas_context.get_preferred_format(
                self._shared.adapter
            )
            # Also configure the canvas
            self._canvas_context.configure(
                device=self._shared.device,
                format=self._target_tex_format,
                usage=wgpu.TextureUsage.RENDER_ATTACHMENT,
            )
        else:
            self._target_tex_format = self._target.format
            # Also enable the texture for render and display usage
            self._target._wgpu_usage |= wgpu.TextureUsage.RENDER_ATTACHMENT
            self._target._wgpu_usage |= wgpu.TextureUsage.TEXTURE_BINDING

        # Prepare render targets.
        self.blend_mode = blend_mode
        self.sort_objects = sort_objects

        # Prepare object that performs the final render step into a texture
        self._flusher = RenderFlusher(self._shared.device)

        # Initialize a small buffer to read pixel info into
        # Make it 256 bytes just in case (for bytes_per_row)
        self._pixel_info_buffer = self._shared.device.create_buffer(
            size=16,
            usage=wgpu.BufferUsage.COPY_DST | wgpu.BufferUsage.MAP_READ,
        )

        if enable_events:
            self.enable_events()

    @property
    def device(self):
        """A reference to the used wgpu device."""
        return self._shared.device

    @property
    def target(self):
        """The render target. Can be a canvas, texture or texture view."""
        return self._target

    @property
    def pixel_ratio(self):
        """The ratio between the number of internal pixels versus the logical pixels on the canvas.

        This can be used to configure the size of the render texture
        relative to the canvas' logical size. By default (value is None) the
        used pixel ratio follows the screens pixel ratio on high-res
        displays, and is 2 otherwise.

        If the used pixel ratio causes the render texture to be larger
        than the physical size of the canvas, SSAA is applied, resulting
        in a smoother final image with less jagged edges. Alternatively,
        this value can be set to e.g. 0.5 to lower* the resolution (e.g.
        for performance during interaction).
        """
        return self._pixel_ratio

    @pixel_ratio.setter
    def pixel_ratio(self, value):
        if value is None:
            self._pixel_ratio = None
        elif isinstance(value, (int, float)):
            self._pixel_ratio = None if value <= 0 else float(value)
        else:
            raise TypeError(
                f"Rendered.pixel_ratio expected None or number, not {value}"
            )

    @property
    def rect(self):
        """The rectangular viewport for the renderer area."""
        return (0, 0) + self.logical_size

    @property
    def logical_size(self):
        """The size of the render target in logical pixels."""
        target = self._target
        if isinstance(target, wgpu.gui.WgpuCanvasBase):
            return target.get_logical_size()
        elif isinstance(target, Texture):
            return target.size[:2]  # assuming pixel-ratio 1
        elif isinstance(target, TextureView):
            return target.texture.size[:2]  # assuming pixel-ratio 1
        else:
            raise TypeError(f"Unexpected render target {target.__class__.__name__}")

    @property
    def physical_size(self):
        """The physical size of the internal render texture."""

        # Get physical size of the target
        target = self._target
        if isinstance(target, wgpu.gui.WgpuCanvasBase):
            target_psize = target.get_physical_size()
        elif isinstance(target, Texture):
            target_psize = target.size[:2]
        elif isinstance(target, TextureView):
            target_psize = target.texture.size[:2]
        else:
            raise TypeError(f"Unexpected render target {target.__class__.__name__}")

        # Determine the pixel ratio of the render texture
        if self._pixel_ratio:
            pixel_ratio = self._pixel_ratio
        else:
            target_lsize = self.logical_size
            pixel_ratio = target_psize[0] / target_lsize[0]
            if pixel_ratio <= 1:
                pixel_ratio = 2.0  # use 2 on non-hidpi displays

        # Determine the physical size of the internal render textures
        return tuple(max(1, int(pixel_ratio * x)) for x in target_lsize)

    @property
    def blend_mode(self):
        """The method for handling transparency:

        * "default" or None: Select the default: currently this is "ordered2".
        * "opaque": single-pass approach that consider every fragment opaque.
        * "ordered1": single-pass approach that blends fragments (using alpha blending).
          Can only produce correct results if fragments are drawn from back to front.
        * "ordered2": two-pass approach that first processes all opaque fragments and then
          blends transparent fragments (using alpha blending) with depth-write disabled. The
          visual results are usually better than ordered1, but still depend on the drawing order.
        * "weighted": two-pass approach that for order independent transparency,
          using alpha weights.
        * "weighted_depth": two-pass approach for order independent transparency,
          with weights based on alpha and depth (McGuire 2013). Note that the depth
          range affects the (quality of the) visual result.
        * "weighted_plus": three-pass approach for order independent transparency,
          in wich the front-most transparent layer is rendered correctly, while
          transparent layers behind it are blended using alpha weights.
        """
        return self._blend_mode

    @blend_mode.setter
    def blend_mode(self, value):
        # Massage and check the input
        if value is None:
            value = "default"
        value = value.lower()
        if value == "default":
            value = "ordered2"
        # Map string input to a class
        m = {
            "opaque": blender_module.OpaqueFragmentBlender,
            "ordered1": blender_module.Ordered1FragmentBlender,
            "ordered2": blender_module.Ordered2FragmentBlender,
            "weighted": blender_module.WeightedFragmentBlender,
            "weighted_depth": blender_module.WeightedDepthFragmentBlender,
            "weighted_plus": blender_module.WeightedPlusFragmentBlender,
        }
        if value not in m:
            raise ValueError(
                f"Unknown blend_mode '{value}', use any of {set(m.keys())}"
            )
        # Set blender object
        self._blend_mode = value
        self._blender = m[value]()
        # If the blend mode has changed, we may need a new _wobject_pipelines
        self._set_wobject_pipelines()
        # If our target is a canvas, request a new draw
        if isinstance(self._target, wgpu.gui.WgpuCanvasBase):
            self._target.request_draw()

    @property
    def sort_objects(self):
        """Whether to sort world objects before rendering. Default False.

        * ``True``: the render order is defined by 1) the object's ``render_order``
          property; 2) the object's distance to the camera; 3) the position object
          in the scene graph (based on a depth-first search).
        * ``False``: don't sort, the render order is defined by the scene graph alone.
        """
        return self._sort_objects

    @sort_objects.setter
    def sort_objects(self, value):
        self._sort_objects = bool(value)

    def _set_wobject_pipelines(self):
        # Each WorldObject has associated with it a wobject_pipeline:
        # a dict that contains the wgpu pipeline objects. This
        # wobject_pipeline is also associated with the blend_mode,
        # because the blend mode affects the pipelines.
        #
        # Each renderer has ._wobject_pipelines, a dict that maps
        # wobject -> wobject_pipeline. This dict is a WeakKeyDictionary -
        # when the wobject is destroyed, the associated pipeline is
        # collected as well.
        #
        # Renderers with the same blend mode can safely share these
        # wobject_pipeline dicts. Therefore, we make use of a global
        # collection. Since this global collection is a
        # WeakValueDictionary, if all renderes stop using a certain
        # blend mode, the associated pipelines are removed as well.
        #
        # In a diagram:
        #
        # _wobject_pipelines_collection -> _wobject_pipelines -> wobject_pipeline
        #        global                         renderer              wobject
        #   WeakValueDictionary              WeakKeyDictionary         dict

        # Below we set this renderer's _wobject_pipelines. Note that if the
        # blending has changed, we automatically invalidate all "our" pipelines.
        self._wobject_pipelines = WgpuRenderer._wobject_pipelines_collection.setdefault(
            self.blend_mode, weakref.WeakKeyDictionary()
        )

    def render(
        self,
        scene: WorldObject,
        camera: Camera,
        *,
        rect=None,
        viewport=None,
        clear_color=None,
        flush=True,
    ):
        """Render a scene with the specified camera as the viewpoint.

        Parameters:
            scene (WorldObject): The scene to render, a WorldObject that
                optionally has child objects.
            camera (Camera): The camera object to use, which defines the
                viewpoint and view transform.
            rect (tuple, optional): The rectangular region to draw into,
                expressed in logical pixels, a.k.a. the viewport.
            clear_color (bool, optional): Whether to clear the color buffer
                before rendering. By default this is True on the first
                call to ``render()`` after a flush, and False otherwise.
            flush (bool, optional): Whether to flush the rendered result into
                the target (texture or canvas). Default True.
        """
        device = self.device

        now = time.perf_counter()  # noqa
        if self._show_fps:
            if not hasattr(self, "_fps"):
                self._fps = now, now, 1
            elif now > self._fps[0] + 1:
                print(f"FPS: {self._fps[2]/(now - self._fps[0]):0.1f}")
                self._fps = now, now, 1
            else:
                self._fps = self._fps[0], now, self._fps[2] + 1

        # Define whether to clear color.
        if clear_color is None:
            clear_color = self._renders_since_last_flush == 0
        clear_color = bool(clear_color)
        self._renders_since_last_flush += 1

        # We always clear the depth, because each render() should be "self-contained".
        # Any use-cases where you normally would control depth-clearing should
        # be covered by the blender. Also, this way the blender can better re-use internal
        # buffers. The only rule is that the color buffer behaves correctly on multiple renders.

        # todo: also note that the fragment shader is (should be) optional
        #      (e.g. depth only passes like shadow mapping or z prepass)

        # Get size for the render textures
        logical_size = self.logical_size
        physical_size = self.physical_size
        if not all(i > 0 for i in logical_size):
            return
        pixel_ratio = physical_size[1] / logical_size[1]

        # Update the render targets
        self._blender.ensure_target_size(device, physical_size)

        # todo: 11-04-2022 - remove viewport arg and this backwards compat logic in the next release
        if viewport is not None:
            if not hasattr(self, "_warned_for_viewport_arg"):
                self._warned_for_viewport_arg = True
                logger.warning(
                    "render(viewport=xx) is deprecated, use rect=xx instead."
                )
            if not rect:
                rect = viewport

        # Get viewport in physical pixels
        if not rect:
            scene_lsize = logical_size
            scene_psize = physical_size
            physical_viewport = 0, 0, physical_size[0], physical_size[1], 0, 1
        elif len(rect) == 4:
            scene_lsize = rect[2], rect[3]
            physical_viewport = [int(i * pixel_ratio + 0.4999) for i in rect]
            physical_viewport = tuple(physical_viewport) + (0, 1)
            scene_psize = physical_viewport[2], physical_viewport[3]
        else:
            raise ValueError(
                "The viewport rect must be None or 4 elements (x, y, w, h)."
            )

        # Ensure that matrices are up-to-date
        scene.update_matrix_world()
        camera.set_view_size(*scene_lsize)
        camera.update_matrix_world()  # camera may not be a member of the scene
        camera.update_projection_matrix()

        # Update stdinfo uniform buffer object that we'll use during this render call
        self._update_stdinfo_buffer(camera, scene_psize, scene_lsize)

        # Get the list of objects to render, as they appear in the scene graph
        wobject_list = []
        scene.traverse(wobject_list.append, True)

        # Ensure each wobject has pipeline info, and filter objects that we cannot render
        wobject_tuples = []
        any_has_changed = False
        for wobject in wobject_list:
            if not wobject.material:
                continue
            wobject_pipeline, has_changed = ensure_pipeline(self, wobject)
            if wobject_pipeline:
                any_has_changed |= has_changed
                wobject_tuples.append((wobject, wobject_pipeline))

        # Command buffers cannot be reused. If we want some sort of re-use we should
        # look into render bundles. See https://github.com/gfx-rs/wgpu-native/issues/154
        # If we do get this to work, we should trigger a new recording
        # when the wobject's children, visibile, render_order, or render_pass changes.

        # Sort objects
        if self.sort_objects:
            sort_func = _get_sort_function(camera)
            wobject_tuples.sort(key=sort_func)

        # Record the rendering of all world objects, or re-use previous recording
        command_buffers = []
        command_buffers += self._render_recording(
            wobject_tuples, physical_viewport, clear_color
        )
        command_buffers += self._blender.perform_combine_pass(self._shared.device)
        command_buffers

        # Collect commands and submit
        device.queue.submit(command_buffers)

        if flush:
            self.flush()

    def flush(self):
        """Render the result into the target texture view. This method is
        called automatically unless you use ``.render(..., flush=False)``.
        """

        # Note: we could, in theory, allow specifying a custom target here.

        if isinstance(self._target, wgpu.gui.WgpuCanvasBase):
            raw_texture_view = self._canvas_context.get_current_texture()
        else:
            if isinstance(self._target, Texture):
                texture_view = self._target.get_view()
            elif isinstance(self._target, TextureView):
                texture_view = self._target
            update_texture(self._shared.device, texture_view.texture)
            update_texture_view(self._shared.device, texture_view)
            raw_texture_view = texture_view._wgpu_texture_view[1]

        # Reset counter (so we can auto-clear the first next draw)
        self._renders_since_last_flush = 0

        command_buffers = self._flusher.render(
            self._blender.color_view,
            None,
            raw_texture_view,
            self._target_tex_format,
        )
        self.device.queue.submit(command_buffers)

    def _render_recording(
        self,
        wobject_tuples,
        physical_viewport,
        clear_color,
    ):

        # You might think that this is slow for large number of world
        # object. But it is actually pretty good. It does iterate over
        # all world objects, and over stuff in each object. But that's
        # it, really.
        # todo: we may be able to speed this up with render bundles though

        command_encoder = self.device.create_command_encoder()
        blender = self._blender

        # ----- compute pipelines

        compute_pass = command_encoder.begin_compute_pass()

        for wobject, wobject_pipeline in wobject_tuples:
            for pinfo in wobject_pipeline.get("compute_pipelines", ()):
                compute_pass.set_pipeline(pinfo["pipeline"])
                for bind_group_id, bind_group in enumerate(pinfo["bind_groups"]):
                    compute_pass.set_bind_group(
                        bind_group_id, bind_group, [], 0, 999999
                    )
                compute_pass.dispatch(*pinfo["index_args"])

        compute_pass.end_pass()

        # ----- render pipelines

        for pass_index in range(blender.get_pass_count()):

            color_attachments = blender.get_color_attachments(pass_index, clear_color)
            depth_attachment = blender.get_depth_attachment(pass_index)
            render_mask = blender.passes[pass_index].render_mask
            if not color_attachments:
                continue

            render_pass = command_encoder.begin_render_pass(
                color_attachments=color_attachments,
                depth_stencil_attachment={
                    **depth_attachment,
                    "stencil_load_value": wgpu.LoadOp.load,
                    "stencil_store_op": wgpu.StoreOp.store,
                },
                occlusion_query_set=None,
            )
            render_pass.set_viewport(*physical_viewport)

            for wobject, wobject_pipeline in wobject_tuples:
                if not (render_mask & wobject_pipeline["render_mask"]):
                    continue
                for pinfo in wobject_pipeline["render_pipelines"]:
                    render_pass.set_pipeline(pinfo["pipelines"][pass_index])
                    for slot, vbuffer in pinfo["vertex_buffers"].items():
                        render_pass.set_vertex_buffer(
                            slot,
                            vbuffer._wgpu_buffer[1],
                            vbuffer.vertex_byte_range[0],
                            vbuffer.vertex_byte_range[1],
                        )
                    for bind_group_id, bind_group in enumerate(pinfo["bind_groups"]):
                        render_pass.set_bind_group(bind_group_id, bind_group, [], 0, 99)
                    # Draw with or without index buffer
                    if pinfo["index_buffer"] is not None:
                        ibuffer = pinfo["index_buffer"]
                        render_pass.set_index_buffer(ibuffer, 0, ibuffer.size)
                        render_pass.draw_indexed(*pinfo["index_args"])
                    else:
                        render_pass.draw(*pinfo["index_args"])

            render_pass.end_pass()

        return [command_encoder.finish()]

    def _update_stdinfo_buffer(self, camera, physical_size, logical_size):
        # Update the stdinfo buffer's data
        stdinfo_data = self._shared.stdinfo_buffer.data
        stdinfo_data["cam_transform"].flat = camera.matrix_world_inverse.elements
        stdinfo_data["cam_transform_inv"].flat = camera.matrix_world.elements
        stdinfo_data["projection_transform"].flat = camera.projection_matrix.elements
        stdinfo_data[
            "projection_transform_inv"
        ].flat = camera.projection_matrix_inverse.elements
        # stdinfo_data["ndc_to_world"].flat = np.linalg.inv(stdinfo_data["cam_transform"] @ stdinfo_data["projection_transform"])
        stdinfo_data["physical_size"] = physical_size
        stdinfo_data["logical_size"] = logical_size
        stdinfo_data["flipped_winding"] = camera.flips_winding
        # Upload to GPU
        self._shared.stdinfo_buffer.update_range(0, 1)
        update_buffer(self._shared.device, self._shared.stdinfo_buffer)

    # Picking

    def get_pick_info(self, pos):
        """Get information about the given window location. The given
        pos is a 2D point in logical pixels (with the origin at the
        top-left). Returns a dict with fields:

        * "ndc": The position in normalized device coordinates, the 3d element
            being the depth (0..1). Can be translated to the position
            in world coordinates using the camera transforms.
        * "rgba": The value in the color buffer. All zero's when rendering
          directly to the screen (bypassing post-processing).
        * "world_object": the object at that location (provided that
          the object supports picking).
        * Additional pick info may be available, depending on the type of
          object and its material. See the world-object classes for details.
        """

        # Make pos 0..1, so we can scale it to the render texture
        logical_size = self.logical_size
        float_pos = pos[0] / logical_size[0], pos[1] / logical_size[1]

        # Prevent out of range and picking before first draw
        out_of_range = not (0 <= float_pos[0] <= 1 and 0 <= float_pos[1] <= 1)
        blender_zero_size = self._blender.size == (0, 0)
        if out_of_range or blender_zero_size:
            return {"rgba": Color(0, 0, 0, 0), "world_object": None}

        # Sample
        encoder = self.device.create_command_encoder()
        self._copy_pixel(encoder, self._blender.color_tex, float_pos, 0)
        self._copy_pixel(encoder, self._blender.pick_tex, float_pos, 8)
        queue = self.device.queue
        queue.submit([encoder.finish()])

        # Collect data from the buffer
        data = self._pixel_info_buffer.map_read()
        color = Color(x / 255 for x in tuple(data[0:4].cast("B")))
        pick_value = tuple(data[8:16].cast("Q"))[0]
        wobject_id = pick_value & 1048575  # 2**20-1
        wobject = id_provider.get_object_from_id(wobject_id)
        # Note: the position in world coordinates is not included because
        # it depends on the camera, but we don't "own" the camera.

        info = {
            "rgba": color,
            "world_object": wobject,
        }

        if wobject:
            pick_info = wobject._wgpu_get_pick_info(pick_value)
            info.update(pick_info)
        return info

    def _copy_pixel(self, encoder, render_texture, float_pos, buf_offset):

        # Map position to the texture index
        w, h, d = render_texture.size
        x = max(0, min(w - 1, int(float_pos[0] * w)))
        y = max(0, min(h - 1, int(float_pos[1] * h)))

        # Note: bytes_per_row must be a multiple of 256.
        encoder.copy_texture_to_buffer(
            {
                "texture": render_texture,
                "mip_level": 0,
                "origin": (x, y, 0),
            },
            {
                "buffer": self._pixel_info_buffer,
                "offset": buf_offset,
                "bytes_per_row": 256,  # render_texture.bytes_per_pixel,
                "rows_per_image": 1,
            },
            copy_size=(1, 1, 1),
        )

    def snapshot(self):
        """Create a snapshot of the currently rendered image."""

        # Prepare
        device = self._shared.device
        texture = self._blender.color_tex
        size = texture.size
        bytes_per_pixel = 4

        # Note, with queue.read_texture the bytes_per_row limitation does not apply.
        data = device.queue.read_texture(
            {
                "texture": texture,
                "mip_level": 0,
                "origin": (0, 0, 0),
            },
            {
                "offset": 0,
                "bytes_per_row": bytes_per_pixel * size[0],
                "rows_per_image": size[1],
            },
            size,
        )

        return np.frombuffer(data, np.uint8).reshape(size[1], size[0], 4)

    def request_draw(self, draw_function=None):
        """Forwards a request_draw call to the target canvas. If the renderer's
        target is not a canvas (e.g. a texture) this function does
        nothing.
        """
        request_draw = getattr(self.target, "request_draw", None)
        if request_draw:
            request_draw(draw_function)

    def enable_events(self):
        """Add event handlers for a specific list of events that are generated
        by the canvas. The handler is the ``convert_event`` method in order to
        convert the Wgpu event dicts into Pygfx event objects."""

        # Check for ``add_event_handler`` attribute. Silently 'fail' as it is
        # perfectly fine to pass a texture as a target to the renderer
        if hasattr(self._target, "add_event_handler"):
            self._target.add_event_handler(self.convert_event, *EVENTS_TO_CONVERT)

    def disable_events(self):
        """Remove the event handlers from the canvas."""
        if hasattr(self._target, "remove_event_handler"):
            self._target.remove_event_handler(self.convert_event, *EVENTS_TO_CONVERT)

    def convert_event(self, event: dict):
        """Converts Wgpu event (dict following jupyter_rfb spec) to Pygfx Event object,
        adds picking info and then dispatches the event in the Pygfx event system.
        """
        event_type = event["event_type"]
        target = None
        if "x" in event and "y" in event:
            info = self.get_pick_info((event["x"], event["y"]))
            target = info["world_object"]
            event["pick_info"] = info

        ev = EVENT_TYPE_MAP[event_type](
            type=event_type, **event, target=target, root=self
        )
        self.dispatch_event(ev)


EVENT_TYPE_MAP = {
    "resize": WindowEvent,
    "close": WindowEvent,
    "pointer_down": PointerEvent,
    "pointer_up": PointerEvent,
    "pointer_move": PointerEvent,
    "double_click": PointerEvent,
    "wheel": WheelEvent,
    "key_down": KeyboardEvent,
    "key_up": KeyboardEvent,
}


EVENTS_TO_CONVERT = (
    "key_down",
    "key_up",
    "pointer_down",
    "pointer_move",
    "pointer_up",
    "wheel",
    "close",
    "resize",
)



from ._base import Material
from ..resources import TextureView
from ..utils import unpack_bitfield, Color


class PointsMaterial(Material):
    """The default material used by Points. Renders (antialiased) disks
    of the given size and color.
    """

    uniform_type = dict(
        color="4xf4",
        size="f4",
    )

    def __init__(
        self,
        color=(1, 1, 1, 1),
        size=1,
        vertex_colors=False,
        vertex_sizes=False,
        map=None,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.color = color
        self._vertex_colors = bool(vertex_colors)
        self.map = map
        self.size = size
        self._vertex_sizes = bool(vertex_sizes)

    def _wgpu_get_pick_info(self, pick_value):
        # This should match with the shader
        values = unpack_bitfield(pick_value, wobject_id=20, index=26, x=9, y=9)
        return {
            "vertex_index": values["index"],
            "point_coord": (values["x"] - 256.0, values["y"] - 256.0),
        }

    @property
    def color(self):
        """The color of the points (if map is not set)."""
        return Color(self.uniform_buffer.data["color"])

    @color.setter
    def color(self, color):
        color = Color(color)
        if (color[3] >= 1) != (self.uniform_buffer.data["color"][3] >= 1):
            self._bump_rev()  # rebuild pipeline if this becomes opaque/transparent
        self.uniform_buffer.data["color"] = color
        self.uniform_buffer.update_range(0, 1)

    @property
    def vertex_colors(self):
        """Whether to use the vertex colors provided in the geometry."""
        return self._vertex_colors

    @vertex_colors.setter
    def vertex_colors(self, value):
        value = bool(value)
        if value != self._vertex_colors:
            self._vertex_colors = value
            self._bump_rev()

    @property
    def size(self):
        """The size (diameter) of the points, in logical pixels."""
        return float(self.uniform_buffer.data["size"])

    @size.setter
    def size(self, size):
        self.uniform_buffer.data["size"] = size
        self.uniform_buffer.update_range(0, 1)

    @property
    def vertex_sizes(self):
        """Whether to use the vertex sizes provided in the geometry."""
        return self._vertex_sizes

    @vertex_sizes.setter
    def vertex_sizes(self, value):
        value = bool(value)
        if value != self._vertex_sizes:
            self._vertex_sizes = value
            self._bump_rev()

    @property
    def map(self):
        """The texture map specifying the color for each texture coordinate.
        The dimensionality of the map can be 1D, 2D or 3D, but should match the
        number of columns in the geometry's texcoords.
        """
        return self._map

    @map.setter
    def map(self, map):
        assert map is None or isinstance(map, TextureView)
        self._map = map

    # todo: sizeAttenuation


class GaussianPointsMaterial(PointsMaterial):
    """A material for points, renders Gaussian blobs with a standard
    deviation of 1/6 of the size.
    """


# idea: a MarkerMaterial with more options for the shape, and an edge around the shape.
# Though perhaps such a material should be part of a higher level plotting lib.



from ..linalg import Matrix4, Vector3
from ..objects._base import WorldObject


class Camera(WorldObject):
    """Abstract base camera.

    The purpose of a camera is to define the viewpoint for rendering a scene.
    This viewpoint consists of its position (in the world) and its projection.

    In other words, it covers the projection of world coordinates to
    normalized device coordinates (NDC), by the (inverse of) the
    camera's own world matrix and the camera's projection transform.
    The former represent the camera's position, the latter is specific
    to the type of camera.

    Note that we follow the NDC coordinate system of WGPU, where
    x and y are in the range 0..1, z is in the range 0..1, and (-1, -1, 0)
    represents the bottom left corner.

    """

    def __init__(self):
        super().__init__()

        self.matrix_world_inverse = Matrix4()
        self.projection_matrix = Matrix4()
        self.projection_matrix_inverse = Matrix4()

    def set_view_size(self, width, height):
        # In logical pixels
        pass

    def update_matrix_world(self, *args, **kwargs):
        super().update_matrix_world(*args, **kwargs)
        self.matrix_world_inverse.get_inverse(self.matrix_world)

    def update_projection_matrix(self):
        raise NotImplementedError()

    def show_object(
        self, target: WorldObject, view_dir=(-1, -1, -1), distance_weight=2
    ):
        """Utility function to position and rotate the camera to ensure
        a particular world object is in view.

        Parameters:
            target: WorldObject
                The object to look at
            view_dir: 3-tuple of float
                Look at the object in this direction
            distance_weight: float
                The camera distance to the object's world position is
                its bounding sphere radius multiplied by this weight

        Returns:
            pos: Vector3
                The world coordinate the camera is looking at
        """
        bsphere = target.get_world_bounding_sphere()
        if bsphere is not None:
            pos, distance = Vector3(*bsphere[:3]), bsphere[3]
        else:
            pos = target.get_world_position()
            # whatever it is has no bounding sphere, so we just pick an
            # arbitrary distance
            distance = 100
        distance *= distance_weight
        self.position.copy(pos).add_scaled_vector(
            Vector3(*view_dir).normalize().negate(), distance
        )
        self.look_at(pos)
        return pos

    @property
    def flips_winding(self):
        """Get whether the camera flips any dimensions causing the
        winding of faces to be flipped. Note that if an even number of
        dimensions are flipped, the winding is not affected.
        """
        return False


class NDCCamera(Camera):
    """A Camera operating in NDC coordinates: its projection matrix
    is the identity transform (but its matrix_world can still be set).

    In the NDC coordinate system of WGPU (and pygfx), x and y are in
    the range 0..1, z is in the range 0..1, and (-1, -1, 0) represents
    the bottom left corner."""

    def update_projection_matrix(self):
        eye = 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1
        self.projection_matrix.set(*eye)
        self.projection_matrix_inverse.set(*eye)


class ScreenCoordsCamera(Camera):
    """A Camera operating in screen coordinates. The depth range is the same
    as in NDC (0 to 1).
    """

    def __init__(self):
        super().__init__()
        self._width = 1
        self._height = 1

    def set_view_size(self, width, height):
        self._width = width
        self._height = height

    def update_projection_matrix(self):
        sx, sy, sz = 2 / self._width, 2 / self._height, 1
        dx, dy, dz = -1, -1, 0
        m = sx, 0, 0, dx, 0, sy, 0, dy, 0, 0, sz, dz, 0, 0, 0, 1
        self.projection_matrix.set(*m)
        self.projection_matrix_inverse.get_inverse(self.projection_matrix)



import numpy as np

from .. import Geometry, Line, LineSegmentMaterial
from ..utils import Color


class AxesHelper(Line):
    """An object indicating the axes directions.

    Parameters:
        size (float): The length of the lines (default 1).
        thickness (float): the thickness of the lines (default 2 px).
    """

    def __init__(self, size=1.0, thickness=2):
        positions = np.array(
            [
                [0, 0, 0],
                [1, 0, 0],
                [0, 0, 0],
                [0, 1, 0],
                [0, 0, 0],
                [0, 0, 1],
            ],
            dtype="f4",
        )
        positions *= size

        colors = np.array(
            [
                [1, 0.6, 0, 1],
                [1, 0.6, 0, 1],  # x is orange-ish
                [0.6, 1, 0, 1],
                [0.6, 1, 0, 1],  # y is yellow-ish
                [0, 0.6, 1, 1],
                [0, 0.6, 1, 1],  # z is blue-ish
            ],
            dtype="f4",
        )

        geometry = Geometry(positions=positions, colors=colors)
        material = LineSegmentMaterial(vertex_colors=True, thickness=thickness, aa=True)

        super().__init__(geometry, material)

    def set_colors(self, x, y, z):
        x, y, z = Color(x), Color(y), Color(z)
        self._geometry.colors.data[0] = x
        self._geometry.colors.data[1] = x
        self._geometry.colors.data[2] = y
        self._geometry.colors.data[3] = y
        self._geometry.colors.data[4] = z
        self._geometry.colors.data[5] = z
        self._geometry.colors.update_range(0, self._geometry.colors.nitems)



"""
A transform gizmo to manipulate world objects.
"""

import numpy as np

from ..objects import Line, Mesh
from ..geometries import Geometry, sphere_geometry, cone_geometry, box_geometry
from ..materials import MeshBasicMaterial, LineMaterial
from ..objects import WorldObject
from ..linalg import Vector3, Quaternion
from ..utils.viewport import Viewport


# Colors in hsluv space - https://www.hsluv.org/
# With H: 0/120/240, S: 100, L: 50
WHITE = "#ffffff"
RED = "#ea0064"
GREEN = "#3f8700"
BLUE = "#007eb7"

THICKNESS = 3


class TransformGizmo(WorldObject):
    """
    A gizmo object that can be used to manipulate (i.e. transform) a world object.

    Aguments:
        object (WorldObject): the object to transform with the gizmo.
        screen_size (float): the approximate size of the widget in logical pixels. Default 100.

    To control the Gizmo:
    * Click the center sphere to toggle between object-space, world-space and screen-space.
    * Grab the center sphere for uniform scaling.
    * Grab the cubes for one-directional scaling (only in object-space).
    * Grab the arrows to translate in one direction.
    * Grab the planes to translate in two directions.
    * Grab the spheres to rotate.
    """

    def __init__(self, object=None, screen_size=100):
        super().__init__()

        # We store these as soon as we get a call in ``add_default_event_handlers``
        self._viewport = None
        self._camera = None

        # A dict that stores the state at the start of a drag. Or None.
        self._ref = None

        # The (approximate) size of the gizmo on screen
        self._screen_size = float(screen_size)

        # Init
        self._create_elements()
        self.toggle_mode("object")  # object, world, screen
        self._highlight()
        self.set_object(object)

    def set_object(self, object):
        """Set the WorldObject to control with the gizmo."""
        if object is None or isinstance(object, WorldObject):
            self._object_to_control = object
        else:
            raise ValueError("The object must be None or a WorldObject instance.")

    def toggle_mode(self, mode=None):
        """Toggle between modes. If the mode is omitted, will move to the next mode.
        Must be one of "object", "world", or "screen".
        """
        modes = "object", "world", "screen"
        if not mode:
            mode = {"object": "world", "world": "screen"}.get(self._mode, "object")
        if mode not in modes:
            raise ValueError(f"Invalid mode '{mode}', must be one of {modes}.")
        self._mode = mode
        self._on_mode_switch()  # The gizmo looks a bit different in each mode

    def _create_elements(self):
        """Create lines, handles, etc."""

        # The location of the scale handles, and where the arcs meet the lines.
        self._halfway = halfway = 0.65

        # --- the parts that are at the center

        # Create screen translate handle
        sphere_geo = sphere_geometry(0.07)
        scale_uniform = Mesh(
            sphere_geo,
            MeshBasicMaterial(color=WHITE),
        )
        scale_uniform.scale.set(1.5, 1.5, 1.5)

        # --- the parts that are fully in one dimension

        # Create lines
        line_geo = Geometry(positions=[(0, 0, 0), (halfway, 0, 0)])
        line_x = Line(
            line_geo,
            LineMaterial(thickness=THICKNESS, color=RED),
        )
        line_y = Line(
            line_geo,
            LineMaterial(thickness=THICKNESS, color=GREEN),
        )
        line_z = Line(
            line_geo,
            LineMaterial(thickness=THICKNESS, color=BLUE),
        )

        # Create 1D translate handles
        cone_geo = cone_geometry(0.1, 0.17)
        cone_geo.positions.data[:] = cone_geo.positions.data[:, ::-1]  # xyz->zyx
        translate_x = Mesh(
            cone_geo,
            MeshBasicMaterial(color=RED),
        )
        translate_y = Mesh(
            cone_geo,
            MeshBasicMaterial(color=GREEN),
        )
        translate_z = Mesh(
            cone_geo,
            MeshBasicMaterial(color=BLUE),
        )
        translate_x.position.set(1, 0, 0)
        translate_y.position.set(0, 1, 0)
        translate_z.position.set(0, 0, 1)

        # Create scale handles
        cube_geo = box_geometry(0.1, 0.1, 0.1)
        scale_x = Mesh(
            cube_geo,
            MeshBasicMaterial(color=RED),
        )
        scale_y = Mesh(
            cube_geo,
            MeshBasicMaterial(color=GREEN),
        )
        scale_z = Mesh(
            cube_geo,
            MeshBasicMaterial(color=BLUE),
        )
        scale_x.position.set(halfway, 0, 0)
        scale_y.position.set(0, halfway, 0)
        scale_z.position.set(0, 0, halfway)

        # --- the parts that are in a plane

        # Create arcs
        t = np.linspace(0, np.pi / 2, 64)
        arc_positions = np.stack([0 * t, np.sin(t), np.cos(t)], 1).astype(np.float32)
        arc_geo = Geometry(positions=arc_positions * halfway)
        arc_yz = Line(
            arc_geo,
            LineMaterial(thickness=THICKNESS, color=RED),
        )
        arc_zx = Line(
            arc_geo,
            LineMaterial(thickness=THICKNESS, color=GREEN),
        )
        arc_xy = Line(
            arc_geo,
            LineMaterial(thickness=THICKNESS, color=BLUE),
        )
        arc_zx.scale.y = -1
        arc_xy.scale.z = -1

        # Create in-plane translate handles
        plane_geo = box_geometry(0.01, 0.15, 0.15)
        translate_yz = Mesh(
            plane_geo,
            MeshBasicMaterial(color=RED),
        )
        translate_zx = Mesh(
            plane_geo,
            MeshBasicMaterial(color=GREEN),
        )
        translate_xy = Mesh(
            plane_geo,
            MeshBasicMaterial(color=BLUE),
        )
        inside_arc = 0.4 * halfway
        translate_yz.position.set(0, inside_arc, inside_arc)
        translate_zx.position.set(inside_arc, 0, inside_arc)
        translate_xy.position.set(inside_arc, inside_arc, 0)

        # Create rotation handles
        # These are positioned on each mode switch
        rotate_yz = Mesh(
            sphere_geo,
            MeshBasicMaterial(color=RED),
        )
        rotate_zx = Mesh(
            sphere_geo,
            MeshBasicMaterial(color=GREEN),
        )
        rotate_xy = Mesh(
            sphere_geo,
            MeshBasicMaterial(color=BLUE),
        )

        # --- post-process

        # Rotate objects to their correct orientation
        for ob in [line_y, translate_y, scale_y, arc_zx, translate_zx, rotate_zx]:
            ob.rotation.set_from_axis_angle(Vector3(0, 0, 1), np.pi / 2)
        for ob in [line_z, translate_z, scale_z, arc_xy, translate_xy, rotate_xy]:
            ob.rotation.set_from_axis_angle(Vector3(0, -1, 0), np.pi / 2)

        # Store the objectss
        self._center_sphere = scale_uniform
        self._line_children = line_x, line_y, line_z
        self._arc_children = arc_yz, arc_zx, arc_xy
        self._translate1_children = translate_x, translate_y, translate_z
        self._translate2_children = translate_yz, translate_zx, translate_xy
        self._translate_children = self._translate1_children + self._translate2_children
        self._scale_children = scale_x, scale_y, scale_z
        self._rotate_children = rotate_yz, rotate_zx, rotate_xy

        # Lines and arcs are never apaque. That way they're also not
        # pickable, so they won't be "in the way".
        for ob in self._line_children + self._arc_children:
            ob.material.opacity = 0.6

        # Assign dimensions
        scale_uniform.dim = None
        for triplet in [
            self._line_children,
            self._translate1_children,
            self._scale_children,
            self._rotate_children,
        ]:
            for i, ob in enumerate(triplet):
                ob.dim = i
        for i, ob in enumerate(self._translate2_children):
            dim = [0, 1, 2]
            dim.pop(i)
            ob.dim = tuple(dim)

        # Attach to the gizmo object
        self.add(*self._line_children)
        self.add(*self._arc_children)
        self.add(self._center_sphere)
        self.add(*self._translate_children)
        self.add(*self._scale_children)
        self.add(*self._rotate_children)

    def _on_mode_switch(self):
        """When the mode is changed, some adjustments are made."""

        # Note: the elements are affected by the mode in various ways.
        # Much of that (e.g. visibility) is already handled in the
        # _update_xx functions, so we don't have to do that here.

        # Update position of rotation handles
        rotate_yz, rotate_zx, rotate_xy = self._rotate_children
        halfway = self._halfway
        on_arc = halfway * 2**0.5 / 2
        if self._mode == "screen":
            rotate_yz.position.set(0, on_arc, 0)
            rotate_zx.position.set(on_arc, 0, 0)
            rotate_xy.position.set(on_arc, on_arc, 0)
        else:
            rotate_yz.position.set(0, on_arc, on_arc)
            rotate_zx.position.set(on_arc, 0, on_arc)
            rotate_xy.position.set(on_arc, on_arc, 0)

    def _highlight(self, object=None):
        """Change the appearance during interaction for visual feedback
        on what is being manipulated.
        """
        # Reset thickness of all lines
        for ob in self._line_children + self._arc_children:
            if ob.material.thickness != THICKNESS:
                ob.material.thickness = THICKNESS

        # Set thickness of lines corresponding to the object
        if object:
            lines = []
            dim = object.dim
            if dim is None:  # center handle
                pass
            elif isinstance(dim, int):
                if object in self._rotate_children:
                    lines.append(self._arc_children[dim])
                else:  # translate1 or scale
                    lines.append(self._line_children[dim])
            else:  # translate2
                lines.append(self._line_children[dim[0]])
                lines.append(self._line_children[dim[1]])
            for ob in lines:
                ob.material.thickness = THICKNESS * 1.75

    # %% Updating before each draw

    def update_matrix_world(self, *args, **kwargs):
        """We overload this method, which gets called by the renderer
        just before rendering. This allows us to prep the gizmo just
        in time.
        """
        # Note that we almost always update the transform (scale,
        # rotation, position) which means the matrix changed, and so
        # does the world_matrix of all children. In effect the uniforms
        # of all elements need updating anyway, so any other changes
        # to wobject properties (e.g. visibility) are "free" - no need
        # to only update if it actually changes.
        if not self._object_to_control:
            self.visible = False
        elif self._viewport and self._camera:
            self.visible = True
            self._update_directions()
            self._update_scale()
            self._update_visibility()
        super().update_matrix_world(*args, **kwargs)

    def _update_directions(self):
        """Calculate the x/y/z reference directions, which depend on
        mode and camera. Calculate these for world-space, ndc-space and
        screen-space.
        """

        camera = self._camera
        scene_size = self._viewport.logical_size
        world_pos = self._object_to_control.position
        ndc_pos = world_pos.clone().project(camera)

        # Calculate direction pairs (world_directions, ndc_directions)
        base_directions = Vector3(1, 0, 0), Vector3(0, 1, 0), Vector3(0, 0, 1)
        if self._mode == "object":
            # The world direction is derived from the object
            rot = self._object_to_control.rotation.clone()
            world_directions = [vec.apply_quaternion(rot) for vec in base_directions]
            # Calculate ndc directions from here
            ndc_directions = [
                world_pos.clone().add(vec).project(camera).sub(ndc_pos)
                for vec in world_directions
            ]
        elif self._mode == "world":
            # The world direction is the base direction
            rot = Quaternion()  # null rotation
            world_directions = base_directions
            # Calculate ndc directions from here
            ndc_directions = [
                world_pos.clone().add(vec).project(camera).sub(ndc_pos)
                for vec in world_directions
            ]
        elif self._mode == "screen":
            # The screen direction is the base_direction
            rot = Quaternion().set_from_rotation_matrix(camera.matrix_world)
            screen_directions = [
                vec.multiply_scalar(self._screen_size) for vec in base_directions
            ]
            # Convert to world directions
            ndc_directions = [
                Vector3(vec.x / scene_size[0] * 2, vec.y / scene_size[1] * 2, -vec.z)
                for vec in screen_directions
            ]
            world_directions = [
                ndc_pos.clone().add(vec).unproject(camera).sub(world_pos)
                for vec in ndc_directions
            ]
        else:  # This cannot happen, in theory
            raise RuntimeError(f"Unexpected mode: '{self._mode}'")

        # Calculate screen directions from ndc_directions (also re-calculate for screen mode)
        # These represent how much one "step" moves on screen.
        # Note how for ndc_directions we have a valid z, but for screen_directions z is 0.
        screen_directions = [
            Vector3(vec.x * scene_size[0] / 2, -vec.y * scene_size[1] / 2, 0)
            for vec in ndc_directions
        ]

        # Store direction lists, to be used during a drag operation.
        self._world_directions = world_directions
        self._ndc_directions = ndc_directions
        self._screen_directions = screen_directions

        # Apply rotation
        self.rotation = rot

    def _update_scale(
        self,
    ):
        """Update the scale of the gizmo so it gets the correct
        (approximate) size on screen.
        """

        # During interaction we don't adjust the scale, this way:
        # * During a rotation the gizmo does not flip,
        #   making it easier to see how much was rotated.
        # * During a translation the gizmo keeps its "world size",
        #   so that the perspective helps you see how the gizmo has moved.
        if self._ref:
            return

        camera = self._camera
        scene_size = self._viewport.logical_size
        world_pos = self._object_to_control.position
        ndc_pos = world_pos.clone().project(camera)

        # Get how our direction vectors express on screen
        ndc_sx = self._screen_size * 2 / scene_size[0]
        ndc_sy = self._screen_size * 2 / scene_size[1]
        vec1 = (
            ndc_pos.clone().add(Vector3(ndc_sx, 0, 0)).unproject(camera).sub(world_pos)
        )
        vec2 = (
            ndc_pos.clone().add(Vector3(0, ndc_sy, 0)).unproject(camera).sub(world_pos)
        )
        scale_scalar = 0.5 * (vec1.length() + vec2.length())

        # Determine the scale of this object, so that it gets the intended size on screen.
        scale = [scale_scalar, scale_scalar, scale_scalar]

        # Determine any flips so that the gizmo faces the camera. Note
        # that this checks whether the vector in question points away
        # from the camera.
        # -----------------------#  So on a view like this, where the widget
        #      | g               #  is on the left, the red leg might still
        # r ___|                 #  just point towards the camera, even though
        #     /                  #  it might not "feel" this way because the
        #    b                   #  blue leg may partly obscure elements of the
        # -----------------------#  red leg.
        for dim, vec in enumerate(self._ndc_directions):
            if vec.z > 0:
                scale[dim] = -scale[dim]

        # Apply scale
        self.scale = Vector3(*scale)

    def _update_visibility(self):
        """Depending on the mode and the orientation of the camera,
        some elements are made invisible.
        """

        screen_directions = self._screen_directions

        # The scaled screen direction matches the size of the widget on screen.
        scale_scalar = abs(self.scale.x)
        scaled_screen_directions = [
            vec.clone().multiply_scalar(scale_scalar) for vec in screen_directions
        ]

        # Determine what directions are orthogonal to the view plane
        show_direction = [True, True, True]
        for dim, vec in enumerate(scaled_screen_directions):
            size = (vec.x**2 + vec.y**2) ** 0.5
            show_direction[dim] = size > 30  # in pixels

        # Also determine whether in-plane elements (arcs and translate2 handles) become hard to see
        show_direction2 = [True, True, True]
        for dim, vec in enumerate(screen_directions):
            dims = [(1, 2), (2, 0), (0, 1)][dim]
            vec1 = screen_directions[dims[0]].clone().normalize()
            vec2 = screen_directions[dims[1]].clone().normalize()
            show_direction2[dim] = abs(vec1.dot(vec2)) < 0.9

        if self._mode == "screen":
            # Show x and y lines and translate1 handles
            for dim, visible in enumerate([True, True, False]):
                self._line_children[dim].visible = visible
                self._translate1_children[dim].visible = visible
            # Show only the xy translate2 handle
            for dim, visible in enumerate([False, False, True]):
                self._translate2_children[dim].visible = visible
                self._arc_children[dim].visible = visible
        else:
            # Lines and arcs are always shown
            for dim in (0, 1, 2):
                self._arc_children[dim].visible = True
                self._line_children[dim].visible = True
            # The translate1 and scale handles depend on their angle to the camera
            for dim in (0, 1, 2):
                self._translate1_children[dim].visible = show_direction[dim]
                self._scale_children[dim].visible = show_direction[dim]
            # The translate2 handles depend on two angles to the camera
            for dim in (0, 1, 2):
                dim1, dim2 = [(1, 2), (2, 0), (0, 1)][dim]
                visible = (
                    show_direction[dim1]
                    and show_direction[dim2]
                    and show_direction2[dim]
                )
                self._translate2_children[dim].visible = visible

        # Per-dimension scaling is only possible in object-mode
        if self._mode != "object":
            for ob in self._scale_children[:3]:
                ob.visible = False

    # %% Event handling

    def add_default_event_handlers(self, viewport, camera):
        # Store objects that we need outside the event handling. In
        # contrast to e.g. a controller, the Gizmo also needs to do
        # some calculation at draw time (or right before a draw, to be
        # precise), and for these calculations it needs the viewport
        # cand camera.
        viewport = Viewport.from_viewport_or_renderer(viewport)
        self._viewport = viewport
        self._camera = camera

        self.add_event_handler(
            self.process_event, "pointer_down", "pointer_move", "pointer_up", "wheel"
        )

    def process_event(self, event):
        """The event handler."""

        # No interaction if there is no object to control
        if not self._object_to_control:
            return

        # Triage over event type
        type = event.type

        if type == "pointer_down":
            if event.button != 1 or event.modifiers:
                return
            self._ref = None
            # NOTE: I imagine that if multiple tools are active, they
            # each ask for picking info, causing multiple buffer reads
            # for the same location. However, with the new event system
            # this is probably not a problem, when wobjects receive events.
            ob = event.target
            if ob not in self.children:
                return
            # Depending on the object under the pointer, we scale/translate/rotate
            if ob == self._center_sphere:
                self._handle_start("scale", event, ob)
            elif ob in self._translate_children:
                self._handle_start("translate", event, ob)
            elif ob in self._scale_children:
                self._handle_start("scale", event, ob)
            elif ob in self._rotate_children:
                self._handle_start("rotate", event, ob)
            # Highlight the object
            self._highlight(ob)
            self._viewport.renderer.request_draw()
            self.set_pointer_capture(event.pointer_id)

        elif type == "pointer_up":
            if not self._ref:
                return
            if self._ref["dim"] is None and self._ref["maxdist"] < 3:
                self.toggle_mode()  # clicked on the center sphere
            self._ref = None
            # De-highlight the object
            self._highlight()
            self._viewport.renderer.request_draw()

        elif type == "pointer_move":
            if not self._ref:
                return
            # Get how far we've moved from starting point - we have a dead zone
            dist = (
                (event.x - self._ref["event_pos"][0]) ** 2
                + (event.y - self._ref["event_pos"][1]) ** 2
            ) ** 0.5
            self._ref["maxdist"] = max(self._ref["maxdist"], dist)
            # Delegate to the correct handler
            if self._ref["maxdist"] < 3:
                pass
            elif self._ref["kind"] == "translate":
                self._handle_translate_move(event)
            elif self._ref["kind"] == "scale":
                self._handle_scale_move(event)
            elif self._ref["kind"] == "rotate":
                self._handle_rotate_move(event)
            # Keep viz up to date
            self._viewport.renderer.request_draw()

    def _handle_start(self, kind, event, ob):
        """Initiate a drag. We create a snapshot of the relevant state at this point."""
        sign = np.sign
        this_pos = self._object_to_control.position.clone()
        ob_pos = ob.get_world_position().clone()
        self._ref = {
            "kind": kind,
            "event_pos": (event.x, event.y),
            "dim": ob.dim,
            "maxdist": 0,
            # Transform at time of start
            "scale": self._object_to_control.scale.clone(),
            "rot": self._object_to_control.rotation.clone(),
            "world_pos": ob_pos,
            "world_offset": ob_pos.clone().sub(this_pos),
            "ndc_pos": ob_pos.clone().project(self._camera),
            # Gizmo direction state at start-time of drag
            "flips": [sign(self.scale.x), sign(self.scale.y), sign(self.scale.z)],
            "world_directions": [vec.clone() for vec in self._world_directions],
            "ndc_directions": [vec.clone() for vec in self._ndc_directions],
            "screen_directions": [vec.clone() for vec in self._screen_directions],
        }

    def _handle_translate_move(self, event):
        """Translate action, either using a translate1 or translate2 handle."""
        # Get dimensions to translate in
        dim = self._ref["dim"]

        # Get how the mouse has moved
        screen_moved = Vector3(
            event.x - self._ref["event_pos"][0],
            event.y - self._ref["event_pos"][1],
            0,
        )
        scene_size = self._viewport.logical_size
        ndc_moved = screen_moved.clone().multiply(
            Vector3(2 / scene_size[0], -2 / scene_size[1], 0)
        )

        # Init new position
        new_position = self._ref["world_pos"].clone()
        new_position.sub(self._ref["world_offset"])

        if isinstance(dim, int):
            # For 1D movement we can project the screen movement to the world-direction.
            # Sample directions
            world_dir = self._ref["world_directions"][dim]
            ndc_dir = self._ref["ndc_directions"][dim].clone()
            screen_dir = self._ref["screen_directions"][dim].clone()
            # Calculate how many times the screen_dir matches the moved direction
            factor = get_scale_factor(screen_dir, screen_moved)
            # Calculate position by moving ndc_pos in that direction
            ndc_pos = self._ref["ndc_pos"].clone().add_scaled_vector(ndc_dir, factor)
            position = ndc_pos.unproject(self._camera)
            # The found position has roundoff errors, let's align it with the world_dir
            world_move = position.clone().sub(self._ref["world_pos"])
            factor = get_scale_factor(world_dir, world_move)
            new_position.add_scaled_vector(world_move, 1)  # world_dir, factor)
        else:
            # For 2d movement we project the cursor vector onto the plane defined
            # by the two world directions.
            dims = dim  # tuple of 2 ints
            # Get reference world pos and the world vectors. Imagine this a plane.
            world_pos = self._ref["world_pos"].clone()
            world_dir1 = self._ref["world_directions"][dims[0]]
            world_dir2 = self._ref["world_directions"][dims[1]]
            # Get the line (in world coords) to move things to
            ndc_pos = self._ref["ndc_pos"].clone()
            ndc_pos1 = ndc_pos.add_scaled_vector(ndc_moved, 1)
            ndc_pos2 = ndc_pos1.clone().add(Vector3(0, 0, 1))
            cursor_world_pos1 = ndc_pos1.unproject(self._camera)
            cursor_world_pos2 = ndc_pos2.unproject(self._camera)
            # Get where line intersects plane, expressed in factors of the world dirs
            factor1, factor2 = get_line_plane_intersection(
                cursor_world_pos1, cursor_world_pos2, world_pos, world_dir1, world_dir2
            )
            new_position.add_scaled_vector(world_dir1, factor1)
            new_position.add_scaled_vector(world_dir2, factor2)

        # Apply
        self._object_to_control.position = new_position.clone()
        self.position = new_position.clone()

    def _handle_scale_move(self, event):
        """Scale action."""
        # Get dimension
        dim = self._ref["dim"]

        # Get how the mouse has moved
        screen_moved = Vector3(
            event.x - self._ref["event_pos"][0],
            event.y - self._ref["event_pos"][1],
            0,
        )

        if dim is None:
            # Uniform scale
            ref_vec = Vector3(1, -1, 0)
            npixels = get_scale_factor(ref_vec, screen_moved)
            scale = 2 ** (npixels / 100)
            scale = Vector3(scale, scale, scale)
        else:
            # Get how much the mouse has moved in the ref direction
            screen_dir = self._ref["screen_directions"][dim]
            factor = get_scale_factor(screen_dir, screen_moved)
            factor *= self._ref["flips"][dim]
            npixels = factor * screen_dir.length()
            # Calculate the relative scale
            scale = [1, 1, 1]
            scale[dim] = 2 ** (npixels / 100)
            scale = Vector3(*scale)

        # Apply
        self._object_to_control.scale = self._ref["scale"].clone().multiply(scale)

    def _handle_rotate_move(self, event):
        """Rotate action."""
        # Get dimension around which to rotate, and the *other* dimensions
        dim = self._ref["dim"]
        dims = [(1, 2), (2, 0), (0, 1)][dim]

        # Get how the mouse has moved
        screen_moved = Vector3(
            event.x - self._ref["event_pos"][0],
            event.y - self._ref["event_pos"][1],
            0,
        )

        # Calculate axis of rotation
        world_dir = self._ref["world_directions"][dim]
        axis = world_dir.clone().normalize()

        # Calculate the vector between the two arrows that span the arc.
        # We need to flip the sign in the right places to make this work.
        screen_dir1 = self._ref["screen_directions"][dims[0]].clone()
        screen_dir2 = self._ref["screen_directions"][dims[1]].clone()
        flip1, flip2 = self._ref["flips"][dims[0]], self._ref["flips"][dims[1]]
        screen_dir1.multiply_scalar(flip1)
        screen_dir2.multiply_scalar(flip2)
        screen_vec = screen_dir2.sub(screen_dir1)

        # Now we can calculate how far the mouse moved in *that* direction.
        factor = get_scale_factor(screen_vec, screen_moved)
        factor = factor * flip1 * flip2
        angle = factor * 2

        # Apply
        rot = Quaternion().set_from_axis_angle(axis, angle)
        self._object_to_control.rotation = rot.multiply(self._ref["rot"])


def get_scale_factor(vec1, vec2):
    """Calculate how many times vec2 fits onto vec1. Basically a dot
    product and a division by their norms.
    """
    factor = vec2.clone().normalize().dot(vec1.clone().normalize())
    factor *= vec2.length() / vec1.length()
    return factor


def get_line_plane_intersection(a0, a1, p0, v1, v2):
    """Get the intersection point of a line onto a plane.
    Args a0 and a1 represent two points on a line. Arg p0 is a point
    on a plane, and v1 and v2 two in-plane vectors. The intersection
    is expressed in factors of v1 and v2.
    """
    # Get the vector from a0 to a1
    av = a1.clone().sub(a0)

    # Get how often this vector must be applied from a0 to get to the plane
    # https://en.wikipedia.org/wiki/Line%E2%80%93plane_intersection
    n = v1.clone().normalize().cross(v2.clone().normalize())
    nom = p0.clone().sub(a0).dot(n)
    denom = av.clone().dot(n)
    at = nom / denom

    # So the point where the line intersects the plane is ...
    p1 = a0.clone().add_scaled_vector(av, at)

    # But let's re-express that in a factor of v1 and v2, so that
    # we really only move in these directions.
    v3 = p1.clone().sub(p0)
    return get_scale_factor(v1, v3), get_scale_factor(v2, v3)



"""Provides utilities to deal with color."""

# Possible improvements:
#
# * Support for HSL colorspace:
#   * Accept CSS compatible "hsl(...)" and "hsla(...)"
#   * Color.hsl property
#   * Color.from_hsl() classmethod
# * Support for HSV color space?
# * Support for HSLuv, a colorspace with uniform brightness https://www.hsluv.org/
# * Color.__add__ another color or a scalar.
# * Color.__mul__ another color or a scalar.
# * Color.lerp(color, t) linear interpolate towards other color.
# * Color.lerpHSL(color, t) same as lerp but interpolate in HSL space.
# * Color.lighter(factor) and Color.darker(factort)

# todo Color.__eq__
# todo: Support "rgb(100%, 50%, 0%)"


import ctypes


F4 = ctypes.c_float * 4


def _float_from_css_value(v, i):
    v = v.strip()
    if v.endswith("%"):
        return float(v[:-1]) / 100
    elif i < 3:
        return float(v) / 255
    else:
        return float(v)


class Color:
    """An object representing a color.

    Internally the color is stored using 4 32-bit floats (rgba).
    It can be instantiated in a variety of ways. E.g. by providing
    the color components as values between 0 and 1:

    * `Color(r, g, b, a)` providing rgba values.
    * `Color(r, g, b)` providing rgb, alpha is 1.
    * `Color(gray, a)` grayscale intensity and alpha.
    * `Color(gray)` grayscale intensity.

    The above variations can also be supplied as a single tuple/list,
    or anything that :

    * `Color((r, g, b))`.

    Named colors:

    * `Color("red")` base color names.
    * `Color("cornflowerblue")` CSS color names.
    * `Color("m")` Matlab color chars.

    Hex colors:

    * `Color("#ff0000")` the commmon hex format.
    * `Color("#ff0000ff")` the hex format that includes alpha.
    * `Color("#ff0")` the short form hex format.
    * `Color("#ff0f")` the short form hex format that includes alpha.

    CSS color functions:

    * `Color("rgb(255, 0, 0)")`.
    * `Color("rgba(255, 0, 0, 1.0)")`.

    """

    # Internally, the color is a ctypes float array
    __slots__ = ["_val"]

    def __init__(self, *args):

        if len(args) == 1:
            color = args[0]
            if isinstance(color, (int, float)):
                self._set_from_tuple(args)
            elif isinstance(color, str):
                self._set_from_str(color)
            else:
                # Assume it's an iterable,
                # may raise TypeError 'object is not iterable'
                self._set_from_tuple(color)
        else:
            self._set_from_tuple(args)

    def __repr__(self):
        # A precision of 4 decimals, i.e. 10001 possible values for each color.
        # We truncate zeros, but make sure the value does not end with a dot.
        f = lambda v: f"{v:0.4f}".rstrip("0").ljust(3, "0")  # noqa: stfu
        return f"Color({f(self.r)}, {f(self.g)}, {f(self.b)}, {f(self.a)})"

    def __len__(self):
        return 4

    def __getitem__(self, index):
        return self._val[index]

    def __iter__(self):
        return self.rgba.__iter__()

    @property
    def __array_interface__(self):
        # Numpy can wrap our memory in an array without copying
        readonly = True
        ptr = ctypes.addressof(self._val)
        x = dict(version=3, shape=(4,), typestr="<f4", data=(ptr, readonly))
        return x

    def _set_from_rgba(self, r, g, b, a):
        self._val = F4(
            max(0.0, min(1.0, float(r))),
            max(0.0, min(1.0, float(g))),
            max(0.0, min(1.0, float(b))),
            max(0.0, min(1.0, float(a))),
        )

    def _set_from_tuple(self, color):
        color = tuple(float(c) for c in color)
        if len(color) == 4:
            self._set_from_rgba(*color)
        elif len(color) == 3:
            self._set_from_rgba(*color, 1)
        elif len(color) == 2:
            self._set_from_rgba(color[0], color[0], color[0], color[1])
        elif len(color) == 1:
            self._set_from_rgba(color[0], color[0], color[0], 1)
        else:
            raise ValueError(f"Cannot parse color tuple with {len(color)} values")

    def _set_from_str(self, color):
        color = color.lower()
        if color.startswith("#"):
            # A hex number
            if len(color) == 7:  # #rrggbb
                self._set_from_rgba(
                    int(color[1:3], 16) / 255,
                    int(color[3:5], 16) / 255,
                    int(color[5:7], 16) / 255,
                    1,
                )
            elif len(color) == 4:  # #rgb
                self._set_from_rgba(
                    int(color[1], 16) / 15,
                    int(color[2], 16) / 15,
                    int(color[3], 16) / 15,
                    1,
                )
            elif len(color) == 9:  # #rrggbbaa
                self._set_from_rgba(
                    int(color[1:3], 16) / 255,
                    int(color[3:5], 16) / 255,
                    int(color[5:7], 16) / 255,
                    int(color[7:9], 16) / 255,
                )
            elif len(color) == 5:  # #rgba
                self._set_from_rgba(
                    int(color[1], 16) / 15,
                    int(color[2], 16) / 15,
                    int(color[3], 16) / 15,
                    int(color[4], 16) / 15,
                )
            else:
                raise ValueError(
                    f"Expecting 4, 5, 7, or 9 chars in a hex number, got {len(color)}."
                )
        elif color.startswith(("rgb(", "rgba(")):
            # A CSS color 'function'
            parts = color.split("(")[1].split(")")[0].split(",")
            parts = [_float_from_css_value(p, i) for i, p in enumerate(parts)]
            if len(parts) == 3:
                self._set_from_rgba(parts[0], parts[1], parts[2], 1)
            elif len(parts) == 4:
                self._set_from_rgba(parts[0], parts[1], parts[2], parts[3])
            else:
                raise ValueError(
                    f"CSS color {color.split('(')[0]}(..) must have 3 or 4 elements, not {len(parts)} "
                )
        else:
            # Maybe a named color
            try:
                color_int = NAMED_COLORS[color.lower()]
            except KeyError:
                raise ValueError(f"Unknown color: '{color}'") from None
            else:
                self._set_from_str(color_int)

    @property
    def rgba(self):
        """The RGBA tuple (values between 0 and 1)."""
        return self._val[0], self._val[1], self._val[2], self._val[3]

    @property
    def rgb(self):
        """The RGB tuple (values between 0 and 1)."""
        return self._val[0], self._val[1], self._val[2]

    @property
    def r(self):
        """The red value."""
        return self._val[0]

    @property
    def g(self):
        """The green value."""
        return self._val[1]

    @property
    def b(self):
        """The blue value."""
        return self._val[2]

    @property
    def a(self):
        """The alpha (transparency) value."""
        return self._val[3]

    @property
    def gray(self):
        """Return the grayscale intensity."""
        # Calculate with the weighted (a.k.a. luminosity) method, same as Matlab
        r, g, b = self.rgb
        return 0.2989 * r + 0.5870 * g + 0.1140 * b

    @property
    def hex(self):
        """The CSS hex string, e.g. "#00ff00". The alpha channel is ignored."""
        r = int(self.r * 255 + 0.5)
        b = int(self.b * 255 + 0.5)
        g = int(self.g * 255 + 0.5)
        i = (r << 16) + (g << 8) + b
        return "#" + hex(i)[2:].rjust(6, "0")

    @property
    def hexa(self):
        """The hex string including alpha, e.g. "#00ff00ff"."""
        r = int(self.r * 255 + 0.5)
        b = int(self.b * 255 + 0.5)
        g = int(self.g * 255 + 0.5)
        a = int(self.a * 255 + 0.5)
        i = (r << 24) + (g << 16) + (b << 8) + a
        return "#" + hex(i)[2:].rjust(8, "0")

    @property
    def css(self):
        """The CSS color string, e.g. "rgba(0,255,0,0.5)"."""
        r, g, b, a = self.rgba
        if a == 1:
            return f"rgb({int(255*r+0.5)},{int(255*g+0.5)},{int(255*b+0.5)})"
        else:
            return f"rgba({int(255*r+0.5)},{int(255*g+0.5)},{int(255*b+0.5)},{a:0.3f})"


NAMED_COLORS = {
    # CSS Level 1
    "black": "#000000",
    "silver": "#C0C0C0",
    "gray": "#808080",
    "white": "#FFFFFF",
    "maroon": "#800000",
    "red": "#FF0000",
    "purple": "#800080",
    "fuchsia": "#FF00FF",
    "green": "#008000",
    "lime": "#00FF00",
    "olive": "#808000",
    "yellow": "#FFFF00",
    "navy": "#000080",
    "blue": "#0000FF",
    "teal": "#008080",
    "aqua": "#00FFFF",
    # CSS Level 2
    "orange": "#FFA500",
    # CSS Color Module Level 3
    "aliceblue": "#F0F8FF",
    "antiquewhite": "#FAEBD7",
    "aquamarine": "#7FFFD4",
    "azure": "#F0FFFF",
    "beige": "#F5F5DC",
    "bisque": "#FFE4C4",
    "blanchedalmond": "#FFEBCD",
    "blueviolet": "#8A2BE2",
    "brown": "#A52A2A",
    "burlywood": "#DEB887",
    "cadetblue": "#5F9EA0",
    "chartreuse": "#7FFF00",
    "chocolate": "#D2691E",
    "coral": "#FF7F50",
    "cornflowerblue": "#6495ED",
    "cornsilk": "#FFF8DC",
    "crimson": "#DC143C",
    "cyan": "#00FFFF",
    "aqua": "#00FFFF",
    "darkblue": "#00008B",
    "darkcyan": "#008B8B",
    "darkgoldenrod": "#B8860B",
    "darkgray": "#A9A9A9",
    "darkgreen": "#006400",
    "darkgrey": "#A9A9A9",
    "darkkhaki": "#BDB76B",
    "darkmagenta": "#8B008B",
    "darkolivegreen": "#556B2F",
    "darkorange": "#FF8C00",
    "darkorchid": "#9932CC",
    "darkred": "#8B0000",
    "darksalmon": "#E9967A",
    "darkseagreen": "#8FBC8F",
    "darkslateblue": "#483D8B",
    "darkslategray": "#2F4F4F",
    "darkslategrey": "#2F4F4F",
    "darkturquoise": "#00CED1",
    "darkviolet": "#9400D3",
    "deeppink": "#FF1493",
    "deepskyblue": "#00BFFF",
    "dimgray": "#696969",
    "dimgrey": "#696969",
    "dodgerblue": "#1E90FF",
    "firebrick": "#B22222",
    "floralwhite": "#FFFAF0",
    "forestgreen": "#228B22",
    "gainsboro": "#DCDCDC",
    "ghostwhite": "#F8F8FF",
    "gold": "#FFD700",
    "goldenrod": "#DAA520",
    "greenyellow": "#ADFF2F",
    "grey": "#808080",
    "honeydew": "#F0FFF0",
    "hotpink": "#FF69B4",
    "indianred": "#CD5C5C",
    "indigo": "#4B0082",
    "ivory": "#FFFFF0",
    "khaki": "#F0E68C",
    "lavender": "#E6E6FA",
    "lavenderblush": "#FFF0F5",
    "lawngreen": "#7CFC00",
    "lemonchiffon": "#FFFACD",
    "lightblue": "#ADD8E6",
    "lightcoral": "#F08080",
    "lightcyan": "#E0FFFF",
    "lightgoldenrodyellow": "#FAFAD2",
    "lightgray": "#D3D3D3",
    "lightgreen": "#90EE90",
    "lightgrey": "#D3D3D3",
    "lightpink": "#FFB6C1",
    "lightsalmon": "#FFA07A",
    "lightseagreen": "#20B2AA",
    "lightskyblue": "#87CEFA",
    "lightslategray": "#778899",
    "lightslategrey": "#778899",
    "lightsteelblue": "#B0C4DE",
    "lightyellow": "#FFFFE0",
    "limegreen": "#32CD32",
    "linen": "#FAF0E6",
    "magenta": "#FF00FF",
    "fuchsia": "#FF00FF",
    "mediumaquamarine": "#66CDAA",
    "mediumblue": "#0000CD",
    "mediumorchid": "#BA55D3",
    "mediumpurple": "#9370DB",
    "mediumseagreen": "#3CB371",
    "mediumslateblue": "#7B68EE",
    "mediumspringgreen": "#00FA9A",
    "mediumturquoise": "#48D1CC",
    "mediumvioletred": "#C71585",
    "midnightblue": "#191970",
    "mintcream": "#F5FFFA",
    "mistyrose": "#FFE4E1",
    "moccasin": "#FFE4B5",
    "navajowhite": "#FFDEAD",
    "oldlace": "#FDF5E6",
    "olivedrab": "#6B8E23",
    "orangered": "#FF4500",
    "orchid": "#DA70D6",
    "palegoldenrod": "#EEE8AA",
    "palegreen": "#98FB98",
    "paleturquoise": "#AFEEEE",
    "palevioletred": "#DB7093",
    "papayawhip": "#FFEFD5",
    "peachpuff": "#FFDAB9",
    "peru": "#CD853F",
    "pink": "#FFC0CB",
    "plum": "#DDA0DD",
    "powderblue": "#B0E0E6",
    "rosybrown": "#BC8F8F",
    "royalblue": "#4169E1",
    "saddlebrown": "#8B4513",
    "salmon": "#FA8072",
    "sandybrown": "#F4A460",
    "seagreen": "#2E8B57",
    "seashell": "#FFF5EE",
    "sienna": "#A0522D",
    "skyblue": "#87CEEB",
    "slateblue": "#6A5ACD",
    "slategray": "#708090",
    "slategrey": "#708090",
    "snow": "#FFFAFA",
    "springgreen": "#00FF7F",
    "steelblue": "#4682B4",
    "tan": "#D2B48C",
    "thistle": "#D8BFD8",
    "tomato": "#FF6347",
    "turquoise": "#40E0D0",
    "violet": "#EE82EE",
    "wheat": "#F5DEB3",
    "whitesmoke": "#F5F5F5",
    "yellowgreen": "#9ACD32",
    # CSS Color Module Level 4
    "rebeccapurple": "#663399",
    # Matlab / Matplotlib
    "b": "#0000FF",
    "g": "#00FF00",
    "r": "#FF0000",
    "c": "#00FFFF",
    "m": "#FF00FF",
    "y": "#FFFF00",
    "k": "#000000",
    "w": "#FFFFFF",
}



from typing import Tuple, Union

from ..linalg import Vector3
from ..utils.viewport import Viewport
from ..renderers import Renderer


class Controller:
    """Base camera controller."""

    def update_camera(self, camera: "Camera") -> "Controller":
        """Update the transform of the camera with the internal transform."""
        rot, pos, zoom = self.get_view()
        camera.rotation.copy(rot)
        camera.position.copy(pos)
        camera.zoom = zoom
        return self

    def add_default_event_handlers(
        self, viewport: Union[Viewport, Renderer], camera: "Camera"
    ):
        """Apply the default interaction mechanism to a wgpu autogui canvas.
        Needs either a viewport or renderer, pus the camera.
        """
        viewport = Viewport.from_viewport_or_renderer(viewport)
        viewport.renderer.add_event_handler(
            lambda event: self.handle_event(event, viewport, camera),
            "pointer_down",
            "pointer_move",
            "pointer_up",
            "wheel",
        )

    def handle_event(self, event, viewport, camera):
        raise NotImplementedError()


def get_screen_vectors_in_world_cords(
    center_world: Vector3, scene_size: Tuple[float, float], camera: "Camera"
) -> Tuple[Vector3, Vector3]:
    """Given a reference center location (in 3D world coordinates)
    Get the vectors corresponding to the x and y direction in screen coordinates.
    These vectors are scaled so that they can simply be multiplied with the
    delta x and delta y.
    """
    center = center_world.clone().project(camera)
    pos1 = Vector3(100, 0, center.z).unproject(camera)
    pos2 = Vector3(0, 100, center.z).unproject(camera)
    pos1.multiply_scalar(0.02 / scene_size[0])
    pos2.multiply_scalar(0.02 / scene_size[1])
    return pos1, pos2  # now they're vecs, really



import wgpu  # only for flags/enums

from . import register_wgpu_render_function
from ._shadercomposer import Binding, WorldObjectShader
from ._utils import to_texture_format
from ...objects import Image
from ...materials import ImageBasicMaterial
from ...resources import Texture, TextureView


vertex_and_fragment = wgpu.ShaderStage.VERTEX | wgpu.ShaderStage.FRAGMENT


def handle_colormap(geometry, material, shader):
    if isinstance(material.map, Texture):
        raise TypeError("material.map is a Texture, but must be a TextureView")
    elif not isinstance(material.map, TextureView):
        raise TypeError("material.map must be a TextureView")
    # Dimensionality
    shader["colormap_dim"] = view_dim = material.map.view_dim
    if material.map.view_dim not in ("1d", "2d", "3d"):
        raise ValueError("Unexpected colormap texture dimension")
    # Texture dim matches image channels
    if int(view_dim[0]) != shader["img_nchannels"]:
        raise ValueError(
            f"Image channels {shader['img_nchannels']} does not match material.map {view_dim}"
        )
    # Sampling type
    fmt = to_texture_format(material.map.format)
    if "norm" in fmt or "float" in fmt:
        shader["colormap_format"] = "f32"
    elif "uint" in fmt:
        shader["colormap_format"] = "u32"
    else:
        shader["colormap_format"] = "i32"
    # Channels
    shader["colormap_nchannels"] = len(fmt) - len(fmt.lstrip("rgba"))
    # Return bindinhs
    return [
        Binding("s_colormap", "sampler/filtering", material.map, "FRAGMENT"),
        Binding("t_colormap", "texture/auto", material.map, "FRAGMENT"),
    ]


@register_wgpu_render_function(Image, ImageBasicMaterial)
def image_renderer(render_info):
    """Render function capable of rendering images."""

    wobject = render_info.wobject
    geometry = wobject.geometry
    material = wobject.material  # noqa
    shader = ImageShader(render_info, climcorrection="")

    bindings = [
        Binding("u_stdinfo", "buffer/uniform", render_info.stdinfo_uniform),
        Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
        Binding("u_material", "buffer/uniform", material.uniform_buffer),
    ]

    topology = wgpu.PrimitiveTopology.triangle_strip
    n = 4

    # Collect texture and sampler
    if geometry.grid is None:
        raise ValueError("Image.geometry must have a grid (texture).")
    else:
        if isinstance(geometry.grid, TextureView):
            view = geometry.grid
        elif isinstance(geometry.grid, Texture):
            view = geometry.grid.get_view(filter="linear")
        else:
            raise TypeError("Image.geometry.grid must be a Texture or TextureView")
        if view.view_dim.lower() != "2d":
            raise TypeError("Image.geometry.grid must a 2D texture (view)")
        # Sampling type
        fmt = to_texture_format(geometry.grid.format)
        if "norm" in fmt or "float" in fmt:
            shader["img_format"] = "f32"
            if "unorm" in fmt:
                shader["climcorrection"] = " * 255.0"
            elif "snorm" in fmt:
                shader["climcorrection"] = " * 255.0 - 128.0"
        elif "uint" in fmt:
            shader["img_format"] = "u32"
        else:
            shader["img_format"] = "i32"
        # Channels
        shader["img_nchannels"] = len(fmt) - len(fmt.lstrip("rgba"))

    bindings.append(Binding("s_img", "sampler/filtering", view, "FRAGMENT"))
    bindings.append(Binding("t_img", "texture/auto", view, vertex_and_fragment))

    # If a colormap is applied ...
    if material.map is not None:
        bindings.extend(handle_colormap(geometry, material, shader))

    # Let the shader generate code for our bindings
    for i, binding in enumerate(bindings):
        shader.define_binding(0, i, binding)

    # Get in what passes this needs rendering
    suggested_render_mask = 3
    if material.opacity < 1:
        suggested_render_mask = 2
    if material.map is not None:
        if shader["colormap_nchannels"] in (1, 3):
            suggested_render_mask = 1
    else:
        if shader["img_nchannels"] in (1, 3):
            suggested_render_mask = 1

    # Put it together!
    return [
        {
            "suggested_render_mask": suggested_render_mask,
            "render_shader": shader,
            "primitive_topology": topology,
            "indices": (range(n), range(1)),
            "vertex_buffers": {},
            "bindings0": bindings,
        }
    ]


sampled_value_to_color = """
        fn sampled_value_to_color(value_rgba: vec4<f32>) -> vec4<f32> {

            // Make it the correct dimension
            $$ if img_nchannels == 1
                let value_raw = value_rgba.r;
            $$ elif img_nchannels == 2
                let value_raw = value_rgba.rg;
            $$ elif img_nchannels == 3
                let value_raw = value_rgba.rgb;
            $$ else
                let value_raw = value_rgba.rgba;
            $$ endif

            // Apply contrast limits
            let value_cor = value_raw {{ climcorrection }};
            let value_clim = (value_cor - u_material.clim[0]) / (u_material.clim[1] - u_material.clim[0]);

            // Apply colormap or compose final color
            $$ if colormap_dim
                // In the render function we make sure that colormap_dim matches img_nchannels
                let color = sample_colormap(value_clim);
            $$ else
                $$ if img_nchannels == 1
                    let r = value_clim;
                    let color = vec4<f32>(r, r, r, 1.0);
                $$ elif img_nchannels == 2
                    let color = vec4<f32>(value_clim.rrr, value_raw.g);
                $$ elif img_nchannels == 3
                    let color = vec4<f32>(value_clim.rgb, 1.0);
                $$ else
                    let color = vec4<f32>(value_clim.rgb, value_raw.a);
                $$ endif
            $$ endif

            return color;
        }
"""


class BaseImageShader(WorldObjectShader):
    def image_helpers(self):
        return (
            sampled_value_to_color
            + """
        struct ImGeometry {
            indices: array<i32,6>;
            positions: array<vec3<f32>,4>;
            texcoords: array<vec2<f32>,4>;
        };

        fn get_im_geometry() -> ImGeometry {
            let size = textureDimensions(t_img);
            var geo: ImGeometry;

            geo.indices = array<i32,6>(0, 1, 2,   3, 2, 1);

            let pos1 = vec2<f32>(-0.5);
            let pos2 = vec2<f32>(size.xy) + pos1;
            geo.positions = array<vec3<f32>,4>(
                vec3<f32>(pos2.x, pos1.y, 0.0),
                vec3<f32>(pos2.x, pos2.y, 0.0),
                vec3<f32>(pos1.x, pos1.y, 0.0),
                vec3<f32>(pos1.x, pos2.y, 0.0),
            );

            geo.texcoords = array<vec2<f32>,4>(
                vec2<f32>(1.0, 0.0),
                vec2<f32>(1.0, 1.0),
                vec2<f32>(0.0, 0.0),
                vec2<f32>(0.0, 1.0),
            );

            return geo;
        }

        fn sample_im(texcoord: vec2<f32>, sizef: vec2<f32>) -> vec4<f32> {
            $$ if img_format == 'f32'
                return textureSample(t_img, s_img, texcoord.xy);
            $$ else
                let texcoords_u = vec2<i32>(texcoord.xy * sizef.xy);
                return vec4<f32>(textureLoad(t_img, texcoords_u, 0));
            $$ endif
        }
    """
        )


class ImageShader(BaseImageShader):
    def get_code(self):
        return (
            self.get_definitions()
            + self.common_functions()
            + self.image_helpers()
            + self.vertex_shader()
            + self.fragment_shader()
        )

    def vertex_shader(self):
        return """

        struct VertexInput {
            [[builtin(vertex_index)]] vertex_index : u32;
        };


        [[stage(vertex)]]
        fn vs_main(in: VertexInput) -> Varyings {

            var geo = get_im_geometry();

            // Select what face we're at
            let index = i32(in.vertex_index);
            let i0 = geo.indices[index];

            // Sample position, and convert to world pos, and then to ndc
            let data_pos = vec4<f32>(geo.positions[i0], 1.0);
            let world_pos = u_wobject.world_transform * data_pos;
            let ndc_pos = u_stdinfo.projection_transform * u_stdinfo.cam_transform * world_pos;

            var varyings: Varyings;
            varyings.position = vec4<f32>(ndc_pos);
            varyings.world_pos = vec3<f32>(world_pos.xyz);
            varyings.texcoord = vec2<f32>(geo.texcoords[i0]);
            return varyings;
        }
        """

    def fragment_shader(self):
        return """

        [[stage(fragment)]]
        fn fs_main(varyings: Varyings) -> FragmentOutput {
            let sizef = vec2<f32>(textureDimensions(t_img));
            let value = sample_im(varyings.texcoord.xy, sizef);
            let color = sampled_value_to_color(value);
            let albeido = color.rgb;

            let final_color = vec4<f32>(albeido, color.a * u_material.opacity);

            // Wrap up
            apply_clipping_planes(varyings.world_pos);
            var out = get_fragment_output(varyings.position.z, final_color);

            $$ if write_pick
            // The wobject-id must be 20 bits. In total it must not exceed 64 bits.
            out.pick = (
                pick_pack(u32(u_wobject.id), 20) +
                pick_pack(u32(varyings.texcoord.x * 4194304.0), 22) +
                pick_pack(u32(varyings.texcoord.y * 4194304.0), 22)
            );
            $$ endif

            return out;
        }
        """



import wgpu

from .. import RenderFunctionRegistry


registry = RenderFunctionRegistry()


def register_wgpu_render_function(wobject_cls, material_cls):
    """Decorator to register a WGPU render function."""

    def _register_wgpu_renderer(f):
        registry.register(wobject_cls, material_cls, f)
        return f

    return _register_wgpu_renderer


def to_vertex_format(format):
    """Convert pygfx' own format to the wgpu format."""
    if format in wgpu.VertexFormat:
        return format
    elif format in wgpu.IndexFormat:
        return format

    # Get primitive type
    primitives = {
        "i1": "sint8",
        "u1": "uint8",
        "i2": "sint16",
        "u2": "uint16",
        "i4": "sint32",
        "u4": "uint32",
        "f2": "float16",
        "f4": "float32",
    }
    primitive = primitives[format[-2:]]

    if len(format) == 2:
        return primitive
    elif len(format) == 4 and format[1] == "x":  # e.g. 3xf4
        if format[0] == "1":
            return primitive
        elif format[0] in "234":
            return primitive + "x" + str(format[0])
        raise ValueError(f"Unexpected tuple size in index/vertex format '{format}'")
    else:
        raise ValueError(f"Unexpected length of index/vertex format '{format}'")


def to_texture_format(format):
    """Convert pygfx' own format to the wgpu format."""
    if format in wgpu.TextureFormat:
        return format

    # Note how we use normalized types (float in the shader) where we can,
    # because these types can work with an interpolating sampler.
    primitives = {
        "i1": "8snorm",
        "u1": "8unorm",
        "i2": "16sint",
        "u2": "16uint",
        "i4": "32sint",
        "u4": "32uint",
        "f2": "16float",
        "f4": "32float",
    }
    primitive = primitives[format[-2:]]

    if len(format) == 2:
        return "r" + primitive
    elif len(format) == 4 and format[1] == "x":  # e.g. 3xf4
        if format[0] == "1":
            return "r" + primitive
        elif format[0] == "2":
            return "rg" + primitive
        elif format[0] == "3":
            return "rgb" + primitive
        elif format[0] == "4":
            return "rgba" + primitive
        else:
            raise ValueError(f"Unexpected tuple size in texture format '{format}'")
    else:
        raise ValueError(f"Unexpected length of texture format '{format}'")



from ._base import Material
from ..resources import TextureView
from ..utils import unpack_bitfield
from ..utils.color import Color


class MeshBasicMaterial(Material):
    """A material for drawing geometries in a simple shaded (flat or
    wireframe) way. This material is not affected by lights.
    """

    uniform_type = dict(
        color="4xf4",
        wireframe="f4",
    )

    def __init__(
        self,
        color=(1, 1, 1, 1),
        vertex_colors=False,
        map=None,
        wireframe=False,
        wireframe_thickness=1,
        side="BOTH",
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.color = color
        self._vertex_colors = bool(vertex_colors)
        self.map = map
        self.wireframe = wireframe
        self.wireframe_thickness = wireframe_thickness
        self.side = side

    def _wgpu_get_pick_info(self, pick_value):
        # This should match with the shader
        values = unpack_bitfield(
            pick_value, wobject_id=20, index=26, coord1=6, coord2=6, coord3=6
        )
        return {
            "face_index": values["index"],
            "face_coord": (
                values["coord1"] / 64,
                values["coord2"] / 64,
                values["coord3"] / 64,
            ),
        }

    @property
    def color(self):
        """The uniform color of the mesh, as an rgba tuple.
        This value is ignored if a texture map is used.
        """
        return Color(self.uniform_buffer.data["color"])

    @color.setter
    def color(self, color):
        color = Color(color)
        if (color[3] >= 1) != (self.uniform_buffer.data["color"][3] >= 1):
            self._bump_rev()  # rebuild pipeline if this becomes opaque/transparent
        self.uniform_buffer.data["color"] = color
        self.uniform_buffer.update_range(0, 1)

    @property
    def vertex_colors(self):
        """Whether to use the vertex colors provided in the geometry."""
        return self._vertex_colors

    @vertex_colors.setter
    def vertex_colors(self, value):
        value = bool(value)
        if value != self._vertex_colors:
            self._vertex_colors = value
            self._bump_rev()

    @property
    def map(self):
        """The texture map specifying the color for each texture coordinate.
        The dimensionality of the map can be 1D, 2D or 3D, but should match the
        number of columns in the geometry's texcoords.
        """
        return self._map

    @map.setter
    def map(self, map):
        assert map is None or isinstance(map, TextureView)
        self._map = map

    @property
    def side(self):
        """Defines which side of faces will be rendered: "FRONT", "BACK", or "BOTH".
        By default this is "BOTH". Setting to "FRONT" or "BACK" will only render
        faces from that side, hiding the other. A feature also known as culling.

        Which side of the mesh is the front is determined by the winding of the faces.
        Counter-clockwise (CCW) winding is assumed. If this is not the case,
        adjust your geometry (using e.g. ``np.fliplr()`` on ``geometry.indices``).
        """
        return self._side

    @side.setter
    def side(self, value):
        side = str(value).upper()
        if side in ("FRONT", "BACK", "BOTH"):
            self._side = side
        else:
            raise ValueError(f"Unexpected side: '{value}'")
        self._bump_rev()

    @property
    def wireframe(self):
        """Render geometry as a wireframe. Default is False (i.e. render as polygons)."""
        return self.uniform_buffer.data["wireframe"] > 0

    @wireframe.setter
    def wireframe(self, value):
        is_wiremode = bool(value)
        was_wiremode = self.uniform_buffer.data["wireframe"] > 0
        if was_wiremode == is_wiremode:
            return
        # Set uniform
        # We use a trick to make negative values indicate no-wireframe mode
        thickness = self.wireframe_thickness
        if is_wiremode:
            self.uniform_buffer.data["wireframe"] = thickness
        else:
            self.uniform_buffer.data["wireframe"] = -thickness
        # Trigger a pipleine rebuild if the mode changes
        self.uniform_buffer.update_range(0, 1)
        self._bump_rev()

    @property
    def wireframe_thickness(self):
        """The thickness of the lines when rendering as a wireframe."""
        return abs(float(self.uniform_buffer.data["wireframe"])) or 1

    @wireframe_thickness.setter
    def wireframe_thickness(self, value):
        value = max(0.01, float(value))
        if self.uniform_buffer.data["wireframe"] > 0:
            self.uniform_buffer.data["wireframe"] = value
        else:
            self.uniform_buffer.data["wireframe"] = -value
        self.uniform_buffer.update_range(0, 1)


# todo: MeshLambertMaterial? In ThreeJS this material uses Gouroud shading with the Lambertian light model.


class MeshPhongMaterial(MeshBasicMaterial):
    """A material affected by light, diffuse and with specular
    highlights. This material uses the Blinn-Phong reflectance model.
    If the specular color is turned off, Lambertian shading is obtained.
    """

    # For reference:
    #
    # Lambertion shading, or Lambertian reflection, is a model to
    # calculate the diffuse component of a lit surface. Using this by
    # itself produces a matte look. All the below use a Lambertion term.
    #
    # Gouraud shading means doing the light-math in the vertex shader
    # and interpolating the final color over the face, often resulting
    # in a somewhat "interpolated" look. Back in the day this mattered
    # for performance, but it's silly now.
    #
    # Phong shading means interpolating the normals and doing the
    # light-math for each fragment.
    #
    # The Phong reflection model refers to the combination of ambient,
    # diffuse and specular lights, and the way that these are
    # calculated.
    #
    # The Blinn-Phong reflection model, also called the modified Phong
    # reflection model, is a tweak to how the reflection is calculated,
    # using a halfway factor, that was intended mostly as a performance
    # optimization, but apparently is a more accurate approximation of
    # how light behaves, or so they say.
    #
    # Flat shading refers to using the same color for the whole face.
    # This is what you get if the geometry has indices that do not share
    # vertices. But we can also obtain it by calculating the face normal
    # using derivatives of the world pos.

    uniform_type = dict(
        emissive_color="4xf4",
        shininess="f4",
    )

    def __init__(self, shininess=30, emissive=(0, 0, 0, 0), **kwargs):
        super().__init__(**kwargs)
        self.emissive = emissive
        self.shininess = shininess

    @property
    def emissive(self):
        """The emissive (light) color of the mesh, as an rgba tuple.
        This color is added to the final color and is unaffected by lighting.
        The alpha channel of this color is ignored.
        """
        return self.uniform_buffer.data["emissive_color"]

    @emissive.setter
    def emissive(self, color):
        self.uniform_buffer.data["emissive_color"] = color
        self.uniform_buffer.update_range(0, 1)

    @property
    def shininess(self):
        """How shiny the specular highlight is; a higher value gives a sharper highlight.
        Default is 30.
        """
        return float(self.uniform_buffer.data["shininess"])

    @shininess.setter
    def shininess(self, value):
        self.uniform_buffer.data["shininess"] = value
        self.uniform_buffer.update_range(0, 1)


class MeshFlatMaterial(MeshPhongMaterial):
    """A material that applies lighting per-face (non-interpolated).
    This gives a "pixelated" look, but can also be usefull if one wants
    to show the (size of) the triangle faces. The shading and
    reflectance model is the same as for ``MeshPhongMaterial``.
    """


# todo: MeshStandardMaterial(MeshBasicMaterial):
# A standard physically based material, using Metallic-Roughness workflow.


# todo: MeshToonMaterial(MeshBasicMaterial):
# A cartoon-style mesh material.


class MeshNormalMaterial(MeshBasicMaterial):
    """A material that maps the normal vectors to RGB colors."""


class MeshNormalLinesMaterial(MeshBasicMaterial):
    """A material that shows surface normals as lines sticking out of the mesh."""

    def _wgpu_get_pick_info(self, pick_value):
        return {}  # No picking for normal lines


class MeshSliceMaterial(MeshBasicMaterial):
    """A material that displays a slice of the mesh."""

    uniform_type = dict(
        plane="4xf4",
        thickness="f4",
    )

    def __init__(self, plane=(0, 0, 1, 0), thickness=2.0, **kwargs):
        super().__init__(**kwargs)
        self.plane = plane
        self.thickness = thickness

    @property
    def plane(self):
        """The plane to slice at, represented with 4 floats ``(a, b, c, d)``,
        which make up the equation: ``ax + by + cz + d = 0`` The plane
        definition applies to the world space (of the scene).
        """
        return self.uniform_buffer.data["plane"]

    @plane.setter
    def plane(self, plane):
        self.uniform_buffer.data["plane"] = plane
        self.uniform_buffer.update_range(0, 1)

    @property
    def thickness(self):
        """The thickness of the line to draw the edge of the mesh."""
        return self.uniform_buffer.data["thickness"]

    @thickness.setter
    def thickness(self, thickness):
        self.uniform_buffer.data["thickness"] = thickness
        self.uniform_buffer.update_range(0, 1)



# flake8: noqa

from ._base import WorldObject, id_provider
from ._events import (
    Event,
    EventTarget,
    EventType,
    PointerEvent,
    KeyboardEvent,
    RootEventHandler,
    WheelEvent,
    WindowEvent,
)
from ._more import Group, Scene, Background, Points, Line, Mesh, Image, Volume
from ._instanced import InstancedMesh



import numpy as np

from ._base import Geometry


# Note that we keep this function separate, because its used by other geometry-generating functions.


def generate_plane(width, height, width_segments, height_segments):
    nx, ny = width_segments + 1, height_segments + 1

    x = np.linspace(-width / 2, width / 2, nx, dtype=np.float32)
    y = np.linspace(-height / 2, height / 2, ny, dtype=np.float32)
    xx, yy = np.meshgrid(x, y)
    xx, yy = xx.flatten(), yy.flatten()
    positions = np.column_stack([xx, yy, np.zeros_like(xx)])

    dim = np.array([width, height], dtype=np.float32)
    texcoords = (positions[..., :2] + dim / 2) / dim
    texcoords[..., 1] = 1 - texcoords[..., 1]

    # the amount of vertices
    indices = np.arange(ny * nx, dtype=np.uint32).reshape((ny, nx))
    # for every panel (height_segments, width_segments) there is a quad (2, 3)
    index = np.empty((height_segments, width_segments, 2, 3), dtype=np.uint32)
    # create a grid of initial indices for the panels
    index[:, :, 0, 0] = indices[
        np.arange(height_segments)[:, None], np.arange(width_segments)[None, :]
    ]
    # the remainder of the indices for every panel are relative
    index[:, :, 0, 1] = index[:, :, 0, 0] + 1
    index[:, :, 0, 2] = index[:, :, 0, 0] + nx
    index[:, :, 1, 0] = index[:, :, 0, 0] + nx + 1
    index[:, :, 1, 1] = index[:, :, 1, 0] - 1
    index[:, :, 1, 2] = index[:, :, 1, 0] - nx

    normals = np.tile(np.array([0, 0, 1], dtype=np.float32), (ny * nx, 1))

    return positions, normals, texcoords, index.reshape((-1, 3))


def plane_geometry(width=1, height=1, width_segments=1, height_segments=1):
    """Create a geometry that represents a 2D plane in a 3D world."""

    positions, normals, texcoords, indices = generate_plane(
        width, height, width_segments, height_segments
    )

    return Geometry(
        indices=indices.reshape((-1, 3)),
        positions=positions,
        normals=normals,
        texcoords=texcoords,
    )
